#################################################################
# CADD - A computer aided drug design using Vision
#
##################################################################
# $Header: /opt/cvs/CADD/__init__.py,v 1.4 2011/05/03 17:58:07 nadya Exp $
# $Id: __init__.py,v 1.4 2011/05/03 17:58:07 nadya Exp $

import time
import sys
import getopt
import os
import user
import warnings
import stat
from string import split

from mglutil.util.packageFilePath import setResourceFolder, getResourceFolder 

__version__ = '0.1';
__revision__ = '1';

# add new type and associated directory in workflows/
WFTypes = {
    ' MD analysis'       : 'MDanalysis',
    ' Virtual Screening' : 'virtualScreening',
    ' Visualization'     : 'visualization',
    ' Docking'           : 'docking',
    'User Workflows'     : 'user'
}

def createRoot():
    """ create a root and hide it
"""
    from Tkinter import Tk
    root = Tk()
    root.withdraw()
    return root


def createCADD(root=None, interactive=True):
    """
"""
    from CADD.VPECADD import VisualProgramingEnvironmentCADD
    ed = VisualProgramingEnvironmentCADD(master=root, name='CADD',
                                     withShell= not interactive)
    return ed


def createSplash():
    """
"""
    iconsDir = os.path.join( __path__[0], 'Icons')
    image_dir = os.path.join( iconsDir, 'Images')

    # insert release date here
    version = ' ' + __version__

    copyright = """(c) 1999-2008 Molecular Simulation Laboratory, The Scripps Research Institute
    ALL RIGHTS RESERVED """
    authors = """Authors: Michel F. Sanner, Daniel Stoffler, Guillaume Vareille"""
    icon = os.path.join( iconsDir, 'cadd.png')
    third_party = """"""

    text = 'Python executable   : '+sys.executable+'\n'
    text += 'CADD script               : '+__file__+'\n'
    text += 'MGLTool packages  : '+'\n'
    
    from user import home
    updates_rc_dir = home + os.path.sep + ".mgltools" + os.sep + 'update'
    latest_tested_rc = updates_rc_dir + os.sep + 'latest_tested'
    latest_rc = updates_rc_dir + os.sep + 'latest' #may or may not be tested
    
    if os.path.exists(latest_tested_rc):
        p = open(latest_tested_rc).read()
        sys.path.insert(0, p)
        text += '    tested  : ' + p +'\n'
    
    if os.path.exists(latest_rc):
        p = open(latest_rc).read()
        sys.path.insert(0, p)
        text += '    nightly : ' + p +'\n'
    
    text += version + ': ' + 'ADD PATH HERE'
    path_data = text

    from mglutil.splashregister.splashscreen import SplashScreen
    from mglutil.splashregister.about import About
    about = About(image_dir=image_dir, third_party=third_party,
                  path_data=path_data,
                  title='CADD', version=version, revision=__revision__, 
                  copyright=copyright, authors=authors, icon=icon)
    splash =  SplashScreen(about, noSplash=False)

    return splash, about


def launchVisionToRunNetworkAsApplication(splash=True):
    root = createRoot()
    if splash:
        splash, about = createSplash()
        lSplashVisibleTimeStart = time.clock()  
    ed = createCADD(root, interactive=True)
    ed.sourceFile(resourceFile='_caddrc')
    masterNet = ed.currentNetwork
    masterNet.isApplication = True
    if splash:
        lSplashVisibleTimeEnd = time.clock()
        while lSplashVisibleTimeEnd - lSplashVisibleTimeStart < 2:
            lSplashVisibleTimeEnd = time.clock()
        splash.finish()
    return masterNet


def mainLoopVisionToRunNetworkAsApplication(ed):
    #ed.master.mainloop()
    sys.stdin = sys.__stdin__
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__
    import code
    mod = __import__('__main__')
    try: # hack to really exit code.interact 
        code.interact( 'CADD Interactive Shell', local=mod.__dict__)
    except:
        pass


##################################################################
# Define a bunch of useful methods
##################################################################
def net():
    return ed.currentNetwork


def nodes():
    return ed.currentNetwork.nodes


def run():
    return ed.currentNetwork.run()


def runCADD(*argv, **kw):
    """The main function for running CADD
listOfDemandedOuts = runCADD('--noSplash',
                               '--runAndExit',
                               'Network0_net.py',
                               ins=[('myeval', 'command', 4),],
                               outs=[('myeval', 'result'),]
                               )
ins: list of tuples, each tuple being (node name, input port name, value)
     the input ports must have a widget.
outs: list of tuples, each tuple being (node name, output port name)
listOfDemandedOuts: list of values in the output ports listed in outs
"""

    ##################################################################
    # Parse options
    ##################################################################

    if type(argv) is tuple:
        if len(argv) == 0:
            argv = None
        elif len(argv) == 1:
            argv = argv[0]
            if type(argv) is not list:
                argv = [argv]
        else:
            argv = list(argv)
    if kw.has_key("ownInterpreter"):
        ownInterpreter = kw["ownInterpreter"]
    else:
        if argv is None:
            argv = ['CADD/bin/runCADD.py']
            ownInterpreter = False
        elif argv[0].endswith('runCADD.py') is False:
            argv.insert(0,'CADD/bin/runCADD.py')
            ownInterpreter = False
        else:
            ownInterpreter = True

    help_msg = """usage: %s <options> <filenames>

    Files filenames ending in "net.py" will be laoded as networks in vision.
    Other files will be executed as Python scripts in which "ed" refers
    to the VisualProgrammingEnvironment instance.
    
            -h or --help : print this message
            -s or --noSplash : doesn't show the CADD splash screen (works only if registered)
            -t or --noTerminal : vision provides its own shell (under menu 'Edit') 
                           instead of the terminal
            --resourceFolder foldername  : stores resource file under .name (defaults to .mgltools)    
            -p or --ipython : create an ipython shell instead of a python shell        
            -r or --run  : run the networks on the command line
            -e or --runAndExit : run the networks and exit

            port values can be passed on the command line (but not to macros):
                             Network_net.py nodeName:portName:value

""" % sys.argv[0]

    try:
        optlist, args = getopt.getopt(argv[1:], 'hirestp', [
            'help', 'interactive', 'noTerminal', 'noSplash', 'resourceFolder=', 'run', 'runAndExit','ipython'] )
    except:
        print "Unknown option!"
        print help_msg
        return

    interactive = 1
    noSplash = False
    runNetwork = False
    once = False
    ipython = False
    for opt in optlist:
        if opt[0] in ('-h', '--help'):
            print help_msg
            return
        elif opt[0] in ('-i', '--interactive'):
            warnings.warn('-i (interactive) is default mode, use --noTerminal to disable it',
                          stacklevel=3)
        elif opt[0] in ('-t', '--noTerminal'):
            interactive = 0
        elif opt[0] in ('-s', '--noSplash',):
            noSplash = True
        elif opt[0] in ('--resourceFolder',):
            setResourceFolder(opt[1])
        elif opt[0] in ('-r', '--run'):
            runNetwork = True
        elif opt[0] in ('-e', '--runAndExit'):
            runNetwork = True
            once = True
        elif opt[0] in ('-p', '--ipython'):
            ipython = True

    from Support.path import path_text, release_path
    print 'Run CADD from ', __path__[0]
    
    if globals().has_key('ed') is False:
        root = createRoot()
        #root.withdraw()
        ##################################################################
        # Splash Screen
        ##################################################################
        if noSplash is False:
            splash, about = createSplash()
            lSplashVisibleTimeStart = time.clock()

        ##################################################################
        # Start CADD
        ##################################################################
        ed = createCADD(root, interactive)
        globals()['ed'] = ed

        ##################################################################
        # Workaround: currently, Vision would crash on some Windows2000 and SGI
        # when running multi-threaded. We turn MT off for those platforms
        ##################################################################
        if os.name == 'nt' or sys.platform == 'irix646':
            ed.configure(withThreads=0)

        ##################################################################
        # Source Resource File (if available)
        ##################################################################
        if getResourceFolder() is None:
            ed.loadLibModule('Vision.StandardNodes')
        else:
            ed.sourceFile(resourceFile='_caddrc')
            # complete vpe menu with additionnal fastLibs
            if hasattr(ed, 'updateFastLibs'):
                ed.updateFastLibs()

        ##################################################################
        # Make sure splash is visible long enough before removing it
        ##################################################################
        if noSplash is False:
            lSplashVisibleTimeEnd = time.clock()
            while lSplashVisibleTimeEnd - lSplashVisibleTimeStart < 2:
                lSplashVisibleTimeEnd = time.clock()
            splash.finish()

        ##################################################################
        # Make vision visible 
        #(before the network loads, so the additional windows will load on top)
        ##################################################################
        if once is False:
            root.deiconify()

    else:
        ed = globals()['ed']
        root = ed.master
        root.master.master.deiconify()

    ##################################################################
    # Load network(s) if specified at startup
    ##################################################################
    lResults = []
    if len(args):
        ed.root.update() # make sure the canvas is shown b4 loading any network
        lArgIndex = 0
        while lArgIndex < len(args):
            if args[lArgIndex][-6:]=='net.py':
                ed.loadNetwork(args[lArgIndex], ins=kw.get('ins'))
                lNodePortValues = []
                lArgIndex += 1
                while lArgIndex < len(args) and args[lArgIndex][-3:]!='.py':
                    lNodePortValues.append(args[lArgIndex])
                    lArgIndex += 1
                ed.currentNetwork.setNodePortValues(lNodePortValues)
                if runNetwork:
                    ed.softrunCurrentNet_cb()
            elif args[lArgIndex][-3:]=='.py':
                execfile(netName)
                lArgIndex += 1
            else:
                lArgIndex += 1

        # lets obtain the asked outs values   
        outs=kw.get('outs')
        if outs is not None:
            for lOut in outs:
                lOutFound = False
                for lNodes in ed.currentNetwork.nodes:
                    if lNodes.name == lOut[0]:
                        if lOutFound is True:
                            warnings.warn("found several nodes %s"%lOut[0] )
                        lOuputPort = lNodes.outputPortByName.get(lOut[1])
                        if lOuputPort is None:
                            warnings.warn("node %s doesn't have port %s"%(lOut[0],lOut[1]) )
                        else:
                           lResults.append(lOuputPort.data)
                        lOutFound = True
                if lOutFound is False:
                    warnings.warn("can't find node %s"%lOut[0] )

    if once is True:
        ed.root.quit()
        ed.root.destroy()
        globals().pop('ed')
        return lResults
    ##################################################################
    # Set the Python Shell
    ##################################################################
    elif ownInterpreter is True:
        if interactive:
            sys.stdin = sys.__stdin__
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__

            mod = __import__('__main__')
            if ipython is True:
                try:
                    # create IPython shell
                    from IPython.Shell import _select_shell
                    sh = _select_shell([])(argv=[], user_ns=mod.__dict__)
                    sh.mainloop()
                except:
                    import code
                    try: # hack to really exit code.interact 
                        code.interact( 'CADD Interactive Shell', local=mod.__dict__)
                    except:
                        pass
            else:
                import code
                try: # hack to really exit code.interact 
                    code.interact( 'CADD Interactive Shell', local=mod.__dict__)
                except:
                    pass

        else:
            ed.master.mainloop()
    else:
        return lResults

packageContainsVFCommands = 1
CRITICAL_DEPENDENCIES = ['NetworkEditor', 'mglutil', 'Support', 'Pmw', 'numpy']
NONCRITICAL_DEPENDENCIES = ['magick', 'PIL', 'opengltk', 'DejaVu', 'grail', 'MolKit', 'symserv','Volume', 'Pmv', 'matplotlib', 'pytz', 'AutoDockTools', 'IPython', 'pymedia']



#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

class RaccoonRemoteServerManager(DebugObj):
    """
        Manager for RaccoonSshClient objects.
        The structure of a server is a dictionary item
        containing the following information:

        servers = { 'name' : { 'address': 'xxx',  
                               'user' : 'xxx',    
                               'password': 'xxx',
                               'type'    : 'docking'  # 'md', 'filtering', 'mdprocessing'
                               'type'    : 'ssh'  # 'pbs', 'sge', 'condor'
                               'pubkey_file'  : None,
                              }
    """

    def __init__(self, servers={}, masterpasswd=None):

        self.servers = []

        # masterpasswd should be used to decript 
        # the master password database 
        # (not available in python)
        self.masterpasswd = masterpasswd
        if servers:
            self.initServerDict(servers)
        
    def initServerDict(self, servers={}): 
        """RaccoonSshClient
        initialize the self.servers list
        """
        for s in servers.keys():
            self.addServer(s, servers[s])

    def addServer(self, name, serverinfo = {}, overwrite=False):
        """
        add a new server 'name' to the server dictionary

        the structure of a serverinfo dictionary is :

            server_info = { 'address'   : xxx,
                            'user'      : xxx,
                            'password'  : xxx,
                            'type'    : 'ssh'  # 'pbs', 'sge', 'condor'
                            'pubkey_file'  : None,
                          }
        """
        if name in self.servers.keys() and not overwrite:
            self.dprint('name already in server dbase: change name or check server info.')
            return False
        self.servers[name] = serverinfo
        if not self.servers[name].haskey('ssh'):
            self.servers[s]['ssh'] = RaccoonSshClient( name = s, address = serverinfo['address'],
                  username=serverinfo['user'], password=serverinfo['password'],
                  pkey=serverinfo['pubkey_file'] )
        return True

    def delServer(self, name):
        """ remove a server from the server list 
            if a connection is open, it gets closed first
        """
        self.server[name]['ssh'].closeconnection()
        del self.server[name]

    
    def saveServers(self, fname):
        for s in self.servers.keys():
            data = self.servers[s]
            # encrypt the data HOW? NOT POSSIBLE!

            # json save it.

            # ...

            # profit!
        pass

    def loadServers(self, fname):
        # from file... unencrypted
        pass






#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#!/usr/bin/env python
from numpy import *
import HelperFunctionsN3P as hf

try:
    DEBUG
except:
    DEBUG = False

COVALENT_BOND_DIST = 1.45
COVALENT_BOND_DIST = COVALENT_BOND_DIST **2 # speed up

def getAromaticRes(receptor):
    # TODO add acids and amides?
    aromatic_res = { "PHE": ["A"],
                 "TYR": ["A"],
                 "HIS": ["A", "N", "NA"],
                 "TRP": ["A", "N", "NA"],
                 "ARG": ["N", "C" , "NA"],
                 "DA" : ["A", "NA", "N"],  # DNA
                 "DC" : ["A", "N", "NA"],
                 "DT" : ["A", "N", "NA"],
                 "DG" : ["A", "N", "NA"],                 
                 "A" : ["A", "NA", "N"],   # DNA/RNA
                 "C" : ["A", "N", "NA"],
                 "G" : ["A", "N", "NA"],                 
                 "T" : ["A", "N", "NA"],
                 "U" : ["A", "N", "NA"],
                 }
    residues = {}
    special = [] # must be treated as a ligand part
    for a in receptor:
        a = a.strip()
        rtype = a[16:21].strip().upper() # NOTE: we can recognize only all caps!
        if a.startswith("HETATM"):
            special.append(a)
        elif a.startswith('ATOM') and rtype in aromatic_res.keys():
            if hf.getAtype(a) in aromatic_res[rtype]:
                res_id = hf.pmvAtomStrip(a).rsplit(":",1)[0] # A:GLY48:N -> A:GLY48
                if not res_id in residues.keys():
                    residues[res_id] = []
                residues[res_id].append(a)
    if special and DEBUG:
        print "found potential co-factor: %d"  % len(special)
        for a in special: print a
    return residues, special # residues = { "A:GLY48" : [ atom, atom, atom..]}

def atomsToVec(atoms):
    a1 = hf.atomCoord(atoms[1])
    a2 = hf.atomCoord(atoms[2])
    a3 = hf.atomCoord(atoms[3])
    centroid = hf.avgCoord(atoms)
    pvector = hf.calcPlane(a1, a2, a3)
    v1 = hf.vector(centroid, a1)
    v2 = hf.vector(centroid, a2)
    normal1 = hf.normalize(cross(v1 ,v2))
    normal2 = hf.normalize(cross(v2 ,v1))
    centroid_norm1 = hf.normalize(hf.vector(centroid, normal1))
    centroid_norm2 = hf.normalize(hf.vector(centroid, normal2))
    return {"centroid":centroid,"plane":hf.normalize(pvector),"normal":[normal1, normal2],\
                        'cnormal' : [centroid_norm1, centroid_norm2]}



#def atomsToVec(atoms):
#    # TODO identify if something is usless (centroid/normal,cnormal)
#    a1 = atomCoord(atoms[1])
#    a2 = atomCoord(atoms[2])
#    a3 = atomCoord(atoms[3])
#    centroid = avgCoord(atoms)
#    pvector = calcPlane(a1, a2, a3)
#    v1 = vector(centroid, a1)
#    v2 = vector(centroid, a2)
#    normal1 = normalize(cross(v1 ,v2))
#    normal2 = normalize(cross(v2 ,v1))
#    centroid_norm1 = normalize(vector(centroid, normal1))
#    centroid_norm2 = normalize(vector(centroid, normal2))
#    return {"centroid":centroid,"plane":normalize(pvector),"normal":[normal1, normal2],\
#                        'cnormal' : [centroid_norm1, centroid_norm2]}






def getResAromaticVectors(residues):
    """
    transform PDBQT residues in centroid,plane data

    INPUT:  residues = { "A:GLY48" : [ atom, atom, atom..], ...}
    OUTPUT: residues = { "A:GLY48" : { centroid_a, centroid_v, plane{, ... }
    """
    for res in residues.keys():
        rtype = res.split(":")[1][0:3] # A:GLY48 => GLY
        try:
            if rtype == "PHE" or rtype == "TYR":
                if len(residues[res]) == 6:
                    pass
                else:
                    if DEBUG: print "Warning! Ignoring residue [ %s ] (missing atoms)" % res
            elif rtype == "TRP":
                if len(residues[res]) >= 9:
                    residues[res] = purgeTrp(residues[res])
                else:
                    if DEBUG: print "Warning! Ignoring residue [ %s ] (missing atoms)" % res
            elif rtype == "HIS":
                if len(residues[res]) >= 5:
                    residues[res] = purgeHis(residues[res])
                else:
                    if DEBUG: print "Warning! Ignoring residue [ %s ] (missing atoms)" % res
            elif rtype == "ARG":
                residues[res] = purgeArg(residues[res])
                if not residues[res] and DEBUG: print "Warning! Ignoring residue [ %s ] (missing atoms)" % res
            elif rtype == "DA" or rtype == "A":
                if len(residues[res]) >= 10:
                    residues[res] = purgeAdenine(residues[res])
                else:
                    if DEBUG: print "Warning! Ignoring residue [ %s ] (missing atoms)" % res
            elif rtype == "DC" or rtype == "C":
                if len(residues[res]) >= 6:
                    residues[res] = purgeAdenine(residues[res])
                else:
                    if DEBUG: print "Warning! Ignoring residue [ %s ] (missing atoms)" % res
            elif rtype == "DG" or rtype == "G":
                if len(residues[res]) >= 10:
                    residues[res] = purgeAdenine(residues[res])
                else:
                    if DEBUG: print "Warning! Ignoring residue [ %s ] (missing atoms)" % res
            if residues[res]:
                residues[res] = atomsToVec(residues[res])
            else:
                del residues[res]
        except:
            if DEBUG: print "Fatal error  in processing residue [ %s ]"% res
            del residues[res]
    return residues

def purgeTrp(atoms):
    """try to purge atoms that are not in the ring(s)"""
    for a in atoms:
        found = False
        if getAtype(a) == "N":
            for c in atoms:
                if not c == a and dist(c,a) < COVALENT_BOND_DIST:
                    found = True
            if not found:
                atoms.remove(a)
                return atoms
    if DEBUG: print "Warning! Residue %s appears to be incomplete" % (atoms[0][17:20]+atoms[0][22:26]+atoms[0][21])
    return False

def purgeHis(atoms):
    """try to purge atoms that are not in the ring(s)"""
    for a in atoms:
        if getAtype(a) == "N" or getAtype(a) == "NA":
            found = 0
            for c in atoms:
                if not c == a and dist(c,a) < COVALENT_BOND_DIST:
                    found = 1
                    break
            if not found:
                atoms.remove(a)
                return atoms
    if DEBUG: print "Warning! Residue %s appears to be incomplete" % (atoms[0][17:20]+atoms[0][22:26]+atoms[0][21])
    return False

def purgeArg(atoms):
    for a in atoms:
        if getAtype(a) == "C":
            candidate = [ a ]
            for n in atoms:
                if not n == a and (getAtype(n) == "N" or getAtype(n) == "NA" ) :
                    if dist(n,a) < COVALENT_BOND_DIST:
                        candidate.append(n)
            if len(candidate) == 4:
                return candidate
    if DEBUG: print "Warning! Residue %s appears to be incomplete" % (atoms[0][17:20]+atoms[0][22:26]+atoms[0][21])
    return False


def purgeAdenine(atoms):
    # look for the single out-of-ring amino nitrogen
    # delete it and return the cleaned up plane
    for a in atoms:
        if getAtype(a) == "N":
            found = 0
            for c in atoms:
                if not c == a and dist(c,a) < COVALENT_BOND_DIST:
                    found += 1
            if not found == 2:
                atoms.remove(a)
                return atoms
    if DEBUG: print "Warning! Residue %s appears to be incomplete" % (atoms[0][17:20]+atoms[0][22:26]+atoms[0][21])
    return False

def purgeCytosine(atoms):
    # look for the single out-of-ring amino nitrogen
    # delete it and return the cleaned up plane
    for a in atoms:
        if getAtype(a) == "N":
            found = 0
            for c in atoms:
                if not c == a and dist(c,a) < COVALENT_BOND_DIST:
                    found += 1
            if not found == 2:
                atoms.remove(a)
                return atoms
    if DEBUG: print "Warning! Residue %s appears to be incomplete" % (atoms[0][17:20]+atoms[0][22:26]+atoms[0][21])
    return False

def purgeGuanine(atom):
    # look for the single out-of-ring amino nitrogen
    # delete it and return the cleaned up plane
    for a in atoms:
        if getAtype(a) == "N":
            found = 0
            for c in atoms:
                if not c == a and dist(c,a) < COVALENT_BOND_DIST:
                    found += 1
            if not found == 2:
                atoms.remove(a)
                return atoms
    if DEBUG: print "Warning! Residue %s appears to be incomplete" % (atoms[0][17:20]+atoms[0][22:26]+atoms[0][21])
    return False
 

def findStacking(set1, set2 = None):
    """
    INPUT(set1, set2): scan the sets to find p- and t-stackings
    INPUT(set1): scan the set to find self-stackings (p-, t-)


    P-STACKING

        ---c---
           |
       A   V
       |
    ---c---       




    """
    DEBUG = 0

    MAXDIST_TSTACK = 5.0 # centroids distance (T-stacking)Angstroms 
    MAXDIST_PSTACK = 4.2 # centroids distance (P-stacking) Angstroms
    
    MAXDIST_PSTACK = MAXDIST_PSTACK**2
    MAXDIST_TSTACK = MAXDIST_TSTACK**2

    #MAXDIST = 5.0 # Angstroms
    #MAXDIST = MAXDIST**2

    PTOL = 29.9 # P-stacking plane angle tolerance (degr) TODO test values on a wide set!
    TTOL = 14.9 # T-stacking plane angle tolerance (degr) TODO test values on a wide set!
    PLANE_DIST_TSTACK = 2.5
    t_accepted = []
    p_accepted = []
    if not set2:
        set2 = set1
    for g1 in set1.keys():
        R1 = set1[g1]
        for g2 in set2.keys():
            if not g1 == g2: # this is for intramolecular stackings (set1 = set2)
                R2 = set2[g2]
                if not ([g1,g2] in p_accepted) and not ([g2,g1] in p_accepted):
                    if (not [g1,g2] in t_accepted) and not ([g2,g1] in t_accepted):
                        d = hf.quickdist(R1['centroid'], R2['centroid'])
                        if d <= MAXDIST_TSTACK:
                            #print "GOOD DIST: %2.3f" % sqrt(d),
                            pdist = abs(hf.dot( R1['normal'][0], (R1['centroid']-R2['centroid']) ))
                            pangle = abs(hf.vecAngle( R1['plane'], R2['plane']))
                            #print "DISTANCE: %2.2f <"% math.sqrt(d), math.sqrt(MAXDIST_TSTACK)
                            #print "PTOL", PTOL
                            #print "PDIST:", pdist 
                            #print "PANGLE:", pangle
                            #print "======"
                            #print "PANGLE-180", pangle-180
                            #print abs(pangle-180)<=PTOL or (pangle-PTOL)<=0
                            #print "PAANGLE-180 < PTOL", abs(pangle-180)<=PTOL
                            #print "PANGLE-PTOL", pangle-PTOL
                            #print "PANGLE-PTOL < 0", (pangle-PTOL)<=0
                            #print "\n\n"
                            if d <= MAXDIST_PSTACK and ((abs(pangle-180)<=PTOL) or ((pangle-PTOL)<=0)):
                                p_accepted.append([ g1, g2])
                                if DEBUG: 
                                    print " ==> P-stacking",
                                    print "%s--%s  plane angle: %2.1f | pdist: %2.2f\n\n" % (g1, g2, pangle, pdist), 
                            elif (( abs(pangle-90) <= TTOL) or ( abs(pangle-270)<=TTOL) ) and pdist < PLANE_DIST_TSTACK:
                                t_accepted.append([ g1, g2])
                                print "XXX"
                                if DEBUG:
                                    print "== > T-stacking",
                                    print "%s--%s  plane angle: %2.1f " % (g1, g2, pangle), 
                                    if ( abs(pangle-90) <= TTOL):
                                        print "ANGLE:",abs(pangle-90), "<", TTOL
                                    if ( abs(pangle-270)<=TTOL):
                                        print "ANGLE:", abs(pangle-270), "<", TTOL
                                    print "PLANE:",pdist, "<", PLANE_DIST_TSTACK
    return p_accepted, t_accepted 



def findRings(graph):
    """
    a very simple ring detection algorightm to identify 5, and 6 member aromatic rings

            A    <- head
           / \   
          B   C  <- tier 1
          |   |
          D   E  <- tier 2
           \ /
            F    <- tier 3, ring closure (6 member-ring)

    
            A    <- head 
           / \              
          B   C  <- tier 1
          |   |
          D---E  <- tier 2, ring closure (5-member ring)


            A   <- head 
            |             
            B   <- tier 1
           / \
          C---D <- tier 2  # SPURIOUS, RECOGNIZED BUT NOT COUNTED

              
    """
    # TODO add a planarity check?
    rings5 = []
    rings6 = []
    if DEBUG: print "- starting ring detection..."
    for head in graph.keys():
        tier1 = graph[head]
        tier2 = []
        tier3 = []
        # populate tier2 
        for node1 in tier1:
            for tmp in graph[node1]:
                if not tmp == head and not tmp in tier2 and (not tmp in tier1) :
                    tier2.append(tmp)
        # populate tier3
        for node2 in tier2:
            for tmp in graph[node2]:
                if (not tmp == head) and (not tmp in tier2) and (not tmp in tier1) and (not tmp in tier3):
                    tier3.append(tmp)
        # 6 member rings
        for x in tier3:
            candidate  = []
            for c in tier2:
                if x in graph[c]:
                    if not c in candidate:
                        candidate.append(c)
            if len(candidate) >1:
                r6 = [ head ] 
                r6.append(x)
                r6 += candidate
                for c in candidate:
                    r6 += hf.intersect( graph[head], graph[c])
                r6.sort()
                if not r6 in rings6:
                    rings6.append( r6 )
                    if DEBUG: print "    6member!", r6
                break
        # 5 member rings
        for c1 in tier2:
            for c2 in tier2:
                if not c1 == c2:
                    if (c2 in graph[c1]) and (c1 in graph[c2]):
                        is_3_ring = False
                        for k in graph[c1]:
                            if k in graph[c2]: 
                                is_3_ring =True
                                if DEBUG: print "       [ ...catched a cycle_3... ]"
                                break
                        if not is_3_ring :
                            r5 = [ head ] 
                            r5.append(c1)
                            r5.append(c2)
                            r5 += hf.intersect( graph[head], graph[c1])
                            r5 += hf.intersect( graph[head], graph[c2])
                            r5.sort()
                            if not r5 in rings5:
                                if DEBUG: print "    5member ring!",r5
                                rings5.append(r5)
                        break
    return rings5, rings6


def searchPiGroupsRec(rec):
    # input : rec atom lines (list)
    aromatic_residues, special = getAromaticRes(rec)
    arom_vectors = getResAromaticVectors(aromatic_residues)
    if special and DEBUG:
        print "INFO: found potential aromatic co-factors in the receptor [ %d ]" % len(special)
    return arom_vectors, special


def findLigandRings(atoms):
    matrix = hf.getDistMatrix(atoms)
    bdist = 2.0 # angstrom of bond dinstance
    graph = hf.makeGraph(nodes_list = range(len(matrix)), distance_matrix = matrix, cutoff = bdist**2)
    rings5,rings6 = findRings(graph)
    RING = rings5+rings6
    ligand_rings = {}
    for r in range(len(RING)):
        # res name first...
        
        # get the first atom PDB string in the ring
        atstring = atoms[RING[r][0]][2]
        ringname = hf.pmvAtomStrip(atstring) 
        ringname_short = ringname.rsplit(":",1)[0].strip()
        if ringname_short:
            c = 0
            while ringname_short+":"+str(c) in ligand_rings.keys():
                c+=1
                if c > 20:
                    ringname_short = None
                    break
        if ringname_short:
            lnam = ringname_short+":"+str(c) 
        else:
            lnam = "rng:"+str(r)

        ligand_rings[lnam] = []
        for a in RING[r]:
            ligand_rings[lnam].append(atoms[a][2])
    return ligand_rings

def searchPiGroupsLig(lig):
    # input : ligand atom lines (list)
    atoms = hf.getFlatAtoms(lig, flat_only = True)
    ligand_rings = findLigandRings(atoms)
    if ligand_rings:
        for r in ligand_rings.keys():
            if DEBUG: 
                print "+++"
                for a in ligand_rings[r]: print a[:-1]
                print "+++\n"
            ligand_rings[r] = atomsToVec(ligand_rings[r])
        return ligand_rings
    else:
        return False

def findLigRecPiStack(lig_atoms, rec_atoms = None, rec_vectors = None):
    # get the receptor aromatic vectors
    # if co-factors are present, process them with ligand tools
    """
    - REC: the function can be called with atoms or pre-calculated vectors
        if atoms:
            vectors will be calculated
        else:
            cached vectors will be used (special can be included too!)
    """
    if not rec_vectors:
        rec_vectors, special = searchPiGroupsRec(rec_atoms)
    if special:
        special = searchPiGroupsLig(special)
        if special:
            #rec_vectors += special
            rec_vectors.update(special)
    lig_vectors = searchPiGroupsLig(lig_atoms)
    pstack = []
    tstack = []
    if lig_vectors:
        lps, lts = findStacking(rec_vectors, lig_vectors)
        if DEBUG:
            if lps: print "Found P-stackings :", len(lps)
            if lts: print "Found T-stackings :", len(lts)
        if lps:
            for p in lps:
                res =  p[0]
                rec_centroid = rec_vectors[p[0]]['centroid']
                lig_centroid = lig_vectors[p[1]]['centroid'] 
                pstack.append([res, rec_centroid, lig_centroid]) 
        if lts:
            for p in lts:
                res =  p[0]
                rec_centroid = rec_vectors[p[0]]['centroid']
                lig_centroid = lig_vectors[p[1]]['centroid'] 
                tstack.append([res, rec_centroid, lig_centroid]) 
    return pstack, tstack

##################################################################
if __name__ == "__main__":
    from sys import argv
    import os
    
    # input must be PDBQT

    ligand = hf.getLines(argv[1])
    name = os.path.basename(argv[1])
    name = os.path.splitext(name)[0]

    try:
        if argv[2] =='-d':
            DEBUG=True
    except:
        pass

    lig_atoms = hf.getAtoms(ligand)
    ligand_rings = searchPiGroupsLig(lig_atoms)
    if not ligand_rings:
        print "No rings in the ligand"
        exit()
    else:
        outname = name+"_vector.pdb"
        print "Writing out centroid and normals [ %s ]" % outname

        out = open(outname, 'w')
        for i in ligand_rings:
            print "RING", i
            res = ligand_rings[i]
            for n in res['normal']:
                v = hf.vecSum( res['centroid'], n )
                out.write( hf.makePdb(coord=v)+'\n' )
            out.write( hf.makePdb(coord=res['centroid'])+'\n' )
        out.close()





#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#!/usr/bin/env python


import sys
from DebugTools import DebugObj
import json, tarfile, urllib2
import AutoDockTools.HelperFunctionsN3P as hf
import os
from datetime import datetime

"""
This file provides the classes for creating Raccoon services.

All services are sub-classes of RaccoonService class that
defines the basic functions and attributes a service should have.

NOTE: To be recognized by the RaccoonServer, services must be added 
      to the dictionary at the end of this file
"""




class RaccoonService(DebugObj):
    """ Base class for Raccoon services.
        Provides:

          self._knownkw      : LIST of know keywords for the configuration
                               of the service

          self._requiredkw   : LIST of minimum set of keywords that need to be 
                               defined for the service to be active

          self.ready         : BOOL of the service passed all validation tests
                               every service must have a self.initService() that
                               check for all requirements for this to be true

          self.config        : DICT containing the configuration of the service;
                               keys in self._knownkwd will be used

          self._srvtype      : STR defining the type of server 

          self._validatekw() : parse the config dictionary 
          self._makeConfigText() : generate the config file of the service using the current
                               service configuration
          self.initService() : initialize the server with the provided configuration dictionary

    """
    def __init__(self, server, debug=False):
        DebugObj.__init__(self, debug)
        self.config = { 'file' : None, 'name' : '', 'comment': '',
                    'service' : None, 'validated' : False}
        # hosting Raccoon server obj
        self.server = server
        # service type
        self._srvtype = None
        # processor, provide the .submit() method
        self.processor = None
        # mandatory kw that need to be set for the service to run
        self._requiredkw = [] 
        # storage for unknown kw
        self._unknownkw = []
        # accepted values for kw (can be empty)
        self._legal_values = {} 
        # text config template
        self._config_template = []
        # service status, when True, it can be initialized
        self.ready = False


    def validateConfig(self, config={}):
        """ validate the config dictionary checking that
            all required kw are presents and that returning
            recognized and unknown keywords
        """
        self.dprint("service current name to validate [%s]" % self.config['name'], new=1)
        # always reset always to default values
        self.config['validated'] = False
        req = 0
        status = { 'accepted' : False, 'reason' : None }
        if 'service' in config.keys():
            if not config['service'] == self._srvtype:
                msg = 'service type mismatch: config[%s] service_obj[%s]'
                msg = msg % (config['service'], self._srvtype) 
                status['reason'] = msg
                self.dprint(msg)
                return False
        else:
            msg = 'server type not defined in the config'
            status['reason'] = msg
            self.dprint(msg)
            return status
        del self._unknownkw[:]
        for k in config.keys():
            if k in self.config:
                if k in self._legal_values.keys():
                    test = self._legal_values[k]
                    if isinstance(test, list):
                        valid = config[k] in test
                    elif isinstance(test, type):
                        valid = False
                        try: 
                            config[k] = test(config[k])
                            valid = True
                        except: pass
                    ## if not config[k] in self._legal_values[k]:
                    if not valid:
                        msg = 'illegal value for kw[%s=%s] (allowed: %s)'
                        #print k, config[k], self._legal_values[k]
                        msg = msg % (k, config[k], self._legal_values[k])
                        status['reason'] = msg
                        self.dprint(msg)
                        return status
                self.config[k] = config[k]
                if k in self._requiredkw:
                    req += 1
            else:
                self.dprint('unknown kw[%s=%s]' % (k,config[k]))
                self._unknownkw.append([k,config[k]])
        if req == len(self._requiredkw):
            self.config['validated'] = True
            self.dprint("Service named [%s] is ready." % self.config['name'])
            status['accepted'] = True
        if len(self._unknownkw):
            print "WARNING: unwknown keywords in service config initialization."
            status['reason'] = 'unkown keywords found'
        return status    
 
    def _makeConfigText(self, template=[]):
        """basic function to generate a config file 
           from the service template

           every service should define a self.makeConfig()
           function built on top of this one
           then add the service-specific customizations
        """
        conf = "\n".join(template)
        conf = conf % self.config
        return conf
        
    def writeServiceConfig(self, fname=None, config_txt=None, force = False):
        """ write the config file on the server """
        status = {'success' : False, 'reason' : None }
        self.dprint("called", new=1, args=1)
        fn = 'writeServiceConfig>'
        # test config dictionary
        if not self.config['validated']:
            msg = 'config not validated'
            status['reason'] = msg
            self.dprint(msg)
            print fn, msg
            return status
        config_txt = self.makeConfig()
        # test filename cases
        if not self.config['file'] == None:
            # service loaded from existing config
            self.dprint('config filename already defined [%s]' % self.config['file'])
            # overwriting original service file
            if self.config['file'] == fname:
                self.dprint( "files saved and to save have same name")
                if not force:
                    msg = 'skipping existing filename'
                    status['reason'] = msg
                    self.dprint(msg)
                    return status
                else:
                    msg = 'updating old config file'
                    self.dprint(msg)
                    status['reason'] = msg
            # writing a (likely) duplicate service 
            elif not self.config['file'] == fname:
                self.dprint( ("pontential service duplicate,"
                             "config loaded from existing "
                             "file[%s]") % self.config['file'])
                if not force:
                    msg = 'potential duplicate services:\n- %s\n- %s\nSkipping...' % (self.config['file'], fname)
                    status['reason'] = msg
                    self.dprint(msg)
                    print fn, msg
                    return status
                else:
                    msg = 'potential duplicate service, proceding'
                    status['reason'] = msg
                    self.dprint(msg)
        # none of the above (or forced to believe so...)
        racfname = self.server.getConfigPath() + '/' + fname
        self.dprint("writing config text to [%s]" % racfname)
        report = self.server.ssh.openfile(racfname, 'w')
        if not type(report) == type(()):
            msg = 'error opening file [%s]' % report
            status['reason'] = msg
            self.dprint(msg)
        else:
            fp, sftp = report
            try:
                fp.write(config_txt+'\n')
                fp.close()
                sftp.close()
                status['success'] = True
                self.config['file'] = racfname
            except:
                msg = "error writing file [%s]" % sys.exc_info()[1]
                status['reason'] = msg
                self.dprint(msg)
        return status


class DockingService(RaccoonService):
    """ 

    """

    def __init__(self, server, config={}, debug=False):
        """
        """
        RaccoonService.__init__(self, server, debug)

        # default config file structure for the docking server
        self._config_template = [
            '############# Raccoon service config file #################',
            '# NAME',
            '# The name of the service as it will be shown in the services list',
            'name = "%(name)s"',
            '\n# COMMENT',
            '# contains a one line description of the service',
            '# this is optional, so leave emtpy quotes if nothing is used',
            'comment = "%(comment)s"',
            '\n# SERVICE TYPE',
            '# define the calculation/data type provided',
            'service = docking',
            '\n# ENGINE TYPE',
            '# this must be a recognized Raccoon engine (i.e. autodock, vina)',
            'engine = %(engine)s',
            '\n# BINARY NAME',
            '# command that is going to be used to run the docking',
            '# if full path is specified, it must be accessible from all nodes',
            'command = "%(command)s"',
            '\n# ENGINE VERSION',
            '# this value will be checked by the service, it can be left blank',
            'ver = "%(ver)s"',
            '\n# MULTITHREAD CALCULATIONS',
            '# specify how many threads are going to be used job;',
            '# if nodes have 4 cores, for example multithread should be',
            '# set to 4; if nodes are single core machines, then multithread',
            '# shuld be set to 1',
            '#      1 : no multithread',
            '#     >1 : use specified cores',
            'multithread = %(multithread)s',
#            '\n# OUTPUT DIRECTORY',
#            '# specify a path where the jobs must be written',
#            '# and the calculations are going to be performed',
#            '# if nothing is specified, the jobs directory',
#            '# will be created in the user home',
#            'outdir = "%(outdir)s"',
            ]
        self._srvtype = 'docking'
        self.config['service'] = self._srvtype
        self.config.update({ 'engine' : None, 'command' : None, 'ver' : None, 
                            'multithread' : 1} )
        self._requiredkw += ['engine', 'command']
        self._legal_values.update( { 'engine': ['autodock', 'vina'] ,
                     'multithread' : int } )

        # links used to install the services on the server
        self.data_source = { 'vina' : {'url': 'http://vina.scripps.edu/download/autodock_vina_1_1_2_linux_x86.tgz',
                                'ver': '1.1.2',
                                'binary': ['vina'],
                                'text' : 'vina --version',
                                },
                    'autodock' : {'url': 'http://autodock.scripps.edu/downloads/autodock-registration/tars/dist423/autodocksuite-4.2.3-i86Linux2.tar.gz',
                                  'ver': '4.2.1',
                                  'binary' : ['autodock4', 'autogrid4'],
                                  'test' : 'autodock4 -v',
                                }
                    }


        if config:
            self.initService(config)



    def makeConfig(self):
        """create the config text for saving to files"""
        template = self._config_template[:]
        return self._makeConfigText(template)


    def initService(self, config):
        """initialize the service from a config dictionary:
            - test if the config is valid
            - validate the executable command
            - 
        """
        self.dprint("==============[ config file ]=============", new=1)
        for c, i in config.items():
            self.dprint("%s : [%s]" % (c,i) )
        self.dprint("==========================================")
        self.dprint("attempt to initialize service [%s]" % config['name'], args=1,new=1)
        self.ready = False
        # config validation
        self.validateConfig(config)
        if not self.config['validated']:
            self.dprint("Service *FAILED* config validation")
            return False
        # command validation
        if not self.validateCommand(which=True):
            self.dprint("Service *FAILED* command validation")
            return False
        # test version (not required.. but important)
        if self.config['ver'] == None:
            if not self.getVersion():
                self.dprint("Service *FAILED* version validation")
                return False
        self.ready = True
        self.dprint("Service [%s] is ready." % self.config['name'])


        scheduler = self.server.systemInfo('scheduler')
        engine = self.config['engine']

        if engine == 'vina':
            #if scheduler == 'pbs':
            #    self.processor = VinaPBSgenerator(self, debug=self.debug)
            #elif scheduler == 'sge':
            #    self.processor = None
            self.processor = VinaSchedulerGenerator(self, debug=self.debug)
        elif engine == 'autodock':
            if scheduler == 'pbs':
                self.processor = None
            elif scheduler == 'sge':
                self.processor = None
        return True



    def validateCommand(self, which=True):
        """ test it if the binary is correct and 
            if so, get the version of the binary
        """ 
        if not self.config['validated']:
            print ("validateCommand> the service config has not been",
                   "validated yet ( use self.validateConfig() )")
            self.dprint("config file not validated yet, returning")
            return False   
        if not self.testCommand(which=which):
            self.dprint("the command wasn't found")
            return False
        self.getVersion()
        return True


    def testCommand(self, which=True, inraccoon=True):
        """ test if the command specified in the config can be found
            as full-path or can be found by the UNIX command 'which'.

            By default, the binary will be searched also in the $RACDIR/bin
            directory, that has priority over system-installed one:
                1. fullpath
                2. racdir (-> set the $RACDIR/bin/ fullpath)
                3. which 
        """
        if not self.config['validated']:
            print ("testCommand> the service config has not been",
                   "validated yet ( use self.validateConfig() )")
            self.dprint("config file not validated yet, returning")
            return False    
        found = False
        bincmd = self.config['command']
        # 1 try fullpath
        found = self.server.ssh.findcommand(bincmd, which=False)
        self.dprint("found (no which) : [%s]" % found)
        # 2. try raccoon
        if not found and inraccoon:
            racbinpath = self.server.getBinPath() 
            found = self.server.ssh.findcommand(racbinpath+bincmd, which=False)
            self.dprint("found (Raccoon path[%s]) :[%s]" % (racbinpath, found) )
        # 3. try which (last resort)
        if not found:
            found = self.server.ssh.findcommand(bincmd, which=which)
        # debug stuff
        if found:
            self.dprint("found binary (final) => [%s]" % (found))
            self.config['command'] = found
        else:
            self.dprint("Impossible to find %s binary => [??]" % (bincmd))
        return found

    def getVersion(self):
        """ get the version of the binary depending on the engine
        
            modify it so the parameters necessary for it will be stored
            at the __init__ time
        """
        if not self.config['validated']:
            print ("getVersion> the service config has not been",
                   "validated yet ( use self.validateConfig() )")
            self.dprint("config file not validated yet, returning")
            return
        engine = self.config['engine']
        bincmd = self.config['command']
        ver = self.config['ver']
        if engine == 'vina':
            cmd = """%s --version | head -1 | awk '{print $3}'""" % bincmd
        elif engine == 'autodock':
            cmd = """%s --version | head -1 | awk '{print $2}'""" % bincmd
        else:
            print "getVersion> UNKNOWN ENGINE!"
            return False
        report = self.server.ssh.execmd(cmd)
        if report == False:
            print "getVersion> no connection... returning"
            self.config['ver'] = '(unknown)'
            return False
        stdout, stderr = report
        self.dprint("STDOUT[%s]" % stdout)
        if len(stdout) > 0:
            if stdout[0] == '':
                print "getVersion> WARNING: no version reported by the binary!"
                ver = '(unknown)'
            else:
                ver = stdout[0]
        else:
            print "getVersion> WARNING: no version found due to possible errors (using the config one)"
        if not ver == self.config['ver']:
            self.dprint("*WARNING* version mismatch: binary[%s] | config[%s]" % (ver, self.config['ver']) )
        self.dprint('Updating service config')
        self.config['ver'] = ver
        return True

    def checkEngine(self):
        """ check if the docking engine binary has been installed already"""
        s = self.server
        engine = self.config['engine']
        info = self.data_source[engine] 
        found = 0
        for bin_name in info['binary']:
            fname = s.getBinPath() + bin_name
            self.dprint('Checking if binary [%s] is on the server' % fname)
            if s.ssh.exists(fname):
                self.dprint('Found.')
                found += 1
        return found == len(info['binary'])


    def installEngine(self, url=None, force=False):
        """
        install the docking engine on the server hosting the service
        """
        status = { 'success': False, 'reason' : None }
        if self.config['engine'] == None:
            self.dprint('No engine defined in the service yet')
            status['reason'] = 'no engine defined'
            return
        engine = self.config['engine']
        # checking url or engine
        if url == None:
            url = self.data_source[ engine ]
            self.dprint("No URL specified, using default:[%s]" % url)
        s = self.server
        info = self.data_source[engine]
        # check if the file is already there
        if self.checkEngine():
            if not force:
                self.dprint('Found! aborting installation')
                status['reason'] = 'binaries already installed'
                return status
            else:
                self.dprint('Found, forced requested, overwriting (maybe)...')
                #s.ssh.delfiles(info[fname])
        # guessing the program link
        if not url:
            url = info['url']
            self.dprint("downloading %s [%s]" % (engine, url))
        else:
            self.dprint("downloading from link: [%s]" % (url))
        # downloading
        success, reply = hf.downloadfile(url, localdirectory=s.scratchdir)
        print "RACCSERVC>", success, reply
        if not success:
            self.dprint("Error downloading [%s]" % reply)
            status['reason'] = reply
            return status
        else:
            installer = reply
        # extracting tar file
        result, source_files = hf.tarextract(tarfilename=installer, 
                                    filelist=info['binary'], outdir=s.scratchdir)
        if result:
            dest_files = [ s.getBinPath() + os.path.basename(f) for f in source_files ]
            files = zip(source_files, dest_files)
            problems = s.ssh.putfiles( flist = files, mode=700) # <-executable
            if len(problems):
                err = "Error copying files:\n"+'\n'.join(problems)
                self.dprint('Found problems:\n%s' % '\n'.join(problems) )
                status['reason'] = err
                return status
            # return the fullpath of binaries
            status['success'] = True
            status['reason'] = dest_files
        return status




#### VINA ####
#
# to run a calculation the following files will be needed:
#
# - list_of_ligands
# - receptor file
# - config file
# 

############### AUTODOCK ###############
# to run a calculation the following files will be needed:
#
#           FULL             CACHED
#     -----------------+------------------
#      list of ligands |  list of ligands
#      receptor file   |  map files
#      list of GPF     | 
#      list of DPF     |  list of DPF
#
#   FULL
#  (list)            (list)         
# ligand_01.pdbqt   ligand_01.gpf   ligand_01_receptor.dpf
# ligand_02.pdbqt   ligand_02.gpf   ligand_02_receptor.dpf
# ligand_03.pdbqt   ligand_03.gpf   ligand_03_receptor.dpf
#




    
class VinaSchedulerGenerator(DebugObj):
    """
            service provide binary and multithread info
    
        Variable names for the pre/post processing
        ==========================================

        $DOCKED_LIGAND  : result of the docking
        $STATUS         : log file where calculations info results are stored


        The start() function provides all functionalities to
            - copy local files to the server
            - generate the config file
            - create index and log files
            - generate scripts for submission (with service and scheduler parms)
    """

    # http://www.brown.edu/Departments/CCV/doc/job-arrays
    # http://stackoverflow.com/questions/2053281/how-to-limit-number-of-concurrently-running-pbs-jobs
    # PBS attributes (max job arrays) : http://linux.die.net/man/7/pbs_server_attributes

    def __init__(self, service, scheduler_parms = {}, debug=True): #, config, preprocess = None, postprocess = None) : # , jobseed, ):
        DebugObj.__init__(self, debug=True)
        print "VINASERVICE INIT"
        self.service = service
        self.server = service.server
        self.preprocess = None
        self.postprocess = None

        self.scheduler_parms = {'walltime' : '24:00:00', 'cputime':'24:00:00', 'queue' : ''}
        for k, v in scheduler_parms.items():
            self.scheduler_parms[k] = v
        self.JOBFILE = 'submission%s.j'
        self.MASTERLOG = "scheduler_master.log"
        self.LOGFILE = 'scheduler_job%s.log'
        self.CONFIGFILE = 'docking.conf'
        self.STATUS = 'status.log'       # each job writes a line when ending
        self.SEEDFILE = 'job.index'         # seed file with all the ligands to dock and dirs to create
                                            #   LIGSOURCE:DOCKING_PATH
                                            #   LIGSOURCE:DOCKING_PATH
                                            #   LIGSOURCE:DOCKING_PATH
        self.options = { 'split' : 1000 }

        self.rec_remote = self.flex_remote = None # populated by the engine later
        # XXX TODO XXX
        # - add support for queues
        # - add specification for mail notifications
        # - add support for non-home dirs
        # - use "-N" option for PBS/SGE job naming (then modify collector to wait for name)

        self._pbs_job_template = [
            """#!/bin/bash""",
            """#PBS -S /bin/bash """,                   # set the shell
            # """#PBS -m %(email)s""",                  # user email for notifications # XXX MISSING GUI
            """#PBS -t %(j_start)d-%(j_end)d""",        # jobs range (1...X, because is used to instruct AWK)
            """#PBS -l nodes=1:ppn=%(multithread)d""",  # multithread value here
            """#PBS -l walltime=%(walltime)s""",        # xx:xx:xx
            """#PBS -d %(vsdir)s""",                    # initial array directory
            """#PBS -o %(vsdir)s/%(logfile)s""",        # log file
            """#PBS -j oe """,                          # combine output and error logs in the same file
            """%(queue)s""",                            # specific queue (optional)
            """# input variables""",                    # -where the input data is injected in the script
            """MASTERLOG="%(vsdir)s/%(masterlog)s" """, # masterlog where all PBS log files of this array are combined
            """SEEDFILE="%(seedfile)s" """,             # 'job.index'
            """STATUS="%(vsdir)s/%(status)s" """,       # 'job_completion.log' where to write update status
            """CONFIG="%(config)s" """,                 # vina config file
            """# end input data""",
            """LINE=`awk "NR==${PBS_ARRAYID}" $SEEDFILE`""",
            """LIGSOURCE=`echo $LINE | awk '{split($0,array,":")} END{print array[1]}'`""",
            """DOCKINGPATH=`echo $LINE | awk '{split($0,array,":")} END{print array[2]}'`""",
            """LIGFILE=`basename $LIGSOURCE`""",
            """LIGNAME=`basename $LIGFILE .pdbqt`""",
            """DOCKLOG=${LIGNAME}.log""",
            """DOCKED_LIGAND=${LIGNAME}_out.pdbqt""",
            """mkdir -p $DOCKINGPATH """,
            """cd $DOCKINGPATH """,
            """touch $STATUS""",
            """%(preprocess)s""",                       # fill here with any pre-processing operation
            """%(binary)s --config "$CONFIG" --ligand "$LIGSOURCE" --out "$DOCKED_LIGAND" --log "$DOCKLOG" """, 
            """%(postprocess)s""",                      # append here any post-processing operations
            """echo $DOCKINGPATH/$DOCKED_LIGAND >> $STATUS""", 
            """echo "====================[ $LIGNAME ]============================" >> $MASTERLOG""",
            """cat %(vsdir)s/%(logfile)s-${PBS_ARRAYID} >> $MASTERLOG """,
            """rm  %(vsdir)s/%(logfile)s-${PBS_ARRAYID}""",
            ]

        self._sge_job_template = [
            """#!/bin/bash""",
            #"""#$ -M %(email)s""",                     # user email for notifications XXX missing interface
            """#$ -S /bin/bash """,                     # set the shell
            """#$ -t %(j_start)d-%(j_end)d""",          # jobs range (1...X, because is used to instruct AWK)
            #"""#$ -l slots=%(multithread)d""",         # multithread DISABLE! THIS WORKS ONLY WITH PARALLEL ENVIRONMENTS!
                                                        # to be checked with :
                                                        #    $ qconf -spl
                                                        #
            """#$ -l h_rt=%(walltime)s""",              # xx:xx:xx
            """#$ -wd %(vsdir)s""",                     # initial array directory
            """#$ -o %(vsdir)s/%(logfile)s""",          # log file
            """#$ -j y """,                             # combine output and error logs in the same file
            """%(queue)s""",                            # specific queue (optional)
            #"""#$ -l qname=queueName """,  #_Clem_ queue name this is site specific, on our clusters _must_ be omitted
            """# input variables""",                    # -where the input data is injected in the script
            """MASTERLOG="%(vsdir)s/%(masterlog)s" """,           # masterlog where all PBS log files of this array are combined
            """SEEDFILE="%(seedfile)s" """,             # 'job.index'
            """STATUS="%(vsdir)s/%(status)s" """,       # 'job_completion.log' where to write update status
            """CONFIG="%(config)s" """,                 # vina config file
            """# end input data""",
            """LINE=`awk "NR==${SGE_TASK_ID}" $SEEDFILE`""", 
            """LIGSOURCE=`echo $LINE | awk '{split($0,array,":")} END{print array[1]}'`""",
            """DOCKINGPATH=`echo $LINE | awk '{split($0,array,":")} END{print array[2]}'`""",
            """LIGFILE=`basename $LIGSOURCE`""",
            """LIGNAME=`basename $LIGFILE .pdbqt`""",
            """DOCKLOG=${LIGNAME}.log""",
            """DOCKED_LIGAND=${LIGNAME}_out.pdbqt""",
            """mkdir -p $DOCKINGPATH """,
            """cd $DOCKINGPATH """,
            """touch $STATUS""",
            """%(preprocess)s""",                       # fill here with any pre-processing operation
            """%(binary)s --config "$CONFIG" --ligand "$LIGSOURCE" --out "$DOCKED_LIGAND" --log "$DOCKLOG" """, 
            """%(postprocess)s""",                      # append here any post-processing operations
            """echo $DOCKINGPATH/$DOCKED_LIGAND >> $STATUS""", 
            #"""echo "====================[ $LIGNAME ]============================" >> $MASTERLOG""",
            #"""cat %(vsdir)s/%(logfile)s-${SGE_TASK_ID} >> $MASTERLOG """,
            #"""rm  %(vsdir)s/%(logfile)s-${SGE_TASK_ID}""",
            ]

    def makeVsName(self, recname=None, ligsource=None, tag=None):
        """ generate the vs name
            this function should be called before the start() to
            generate the name to be checked for duplicate experiments
        """
        if tag == None:
            tag = self.jobinfo['tag']
        if recname == None:
            recname = self.recname
        if ligsource == None:
            ligsource = self.ligsource
        self.dprint("recname = [%s]")
        self.dprint("ligsource = ", ligsource)
        self.dprint("tag [%s]"% tag)
        # library name (combining all libraries selected)
        libname = "_".join([ x.name() for x in ligsource])
        libname = hf.validFilename(libname)
        name = "%s--%s" % (recname, libname)
        if tag: name = '%s_%s' % (name, tag)
        self.dprint("generated name [%s]" % name)
        return name

    def start(self, jobinfo, engine, expdir, ligsource, recname, preprocess='', postprocess='', submit=True, callback=None ):
        """ 
            jobinfo     : dictionary with prj and exp names
            engine      : Raccoon engine object
            expdir      : experiment directory (already created)
            ligsource   : list of RaccoonLibrary objects
            recname     : receptor name (to retrieve rec info with the engine)
            preprocess  : preprocessing objects to call for generating the script code 
            postprocess : postprocessing objects to call for generating the script code
                            [ default echoing into the $STATUS]
        """
        # XXX preprocess and postprocess
        #     should become the names of the other services to be called
        #     to generate the text to be added to the script...
        # postprocess = server.service(preprocess)
        # text = postprocess.generate()
        # append text to the service
        self.job_id = []
        self.jobinfo = jobinfo
        self.engine = engine
        self.expdir = expdir
        self.ligsource = ligsource
        self.recname = recname
        self.recinfo = self.engine.RecBook[self.recname]
        if not preprocess == '':
            self.preprocess = preprocess.generate(lig='$LIGSOURCE', rec=self.rec_remote)
        else:
            self.preprocess = ''
        if not postprocess == '':
            self.postprocess = postprocess.generate(lig='$DOCKED_LIGAND', rec=self.rec_remote)
        else:
            self.postprocess = ''
        self.callback = callback

        if self.jobinfo.has_key('name'):
            self.name = self.jobinfo['name']
            self.dprint("got name [%s]" % self.name)
        else:
            self.name = self.makeVsName()
            self.dprint("made name [%s]" % self.name)
        vsdir = '/'.join([self.expdir,self.name]) 
        vsdir = hf.validPath(vsdir)
        self.dprint("VS dir generated [%s]" % vsdir)

        submission_data = {'error' : None , 'info' : None}
        # make vsdir
        if self.callback: callback("Creating VS directories...")
        result = self.server.ssh.makedir(vsdir)
        if not result == True:
            error = 'error creating vsdir[%s] (aborting)' % vsdir
            submission_data['errors'] = error
            self.dprint(error)
            return submission_data
        self.vsdir = self.server.ssh.normpath(vsdir)
        # copy rec files
        if self.callback: callback("Copying receptor files...")
        if not self.copyrecfiles():
            error = "Copy receptor error: aborting"
            submission_data['errors'] = error
            self.dprint(error)
            return submission_data
        if not self.makeliglist():
            error = "Error generating ligand list, aborting"
            self.dprint(error)
            submission_data['errors'] = error
            return submission_data
        # generate config file
        if self.callback: callback("Generating docking config file...")
        if not self.makeconf():
            error = "Generate config: aborting"
            self.dprint(error)
            submission_data['errors'] = error
            return submission_data
        # generate jobs
        if self.callback: callback("Generating PBS jobs...")
        if not self.generatejobs():
            error = "Generating jobs error: aborting"
            self.dprint(error)
            submission_data['errors'] = error
            return submission_data
        if submit:
            if self.callback: callback("Submitting jobs...")
            submission_data = self.submit()
            return submission_data

    def copyrecfiles(self):
        """ transfer local receptor files to the remote server"""
        # copy receptor on vsdir
        self.dprint("copying rec files")
        files = []
        rec_local = self.recinfo['filename']
        self.rec_remote = self.vsdir + '/' + os.path.basename(rec_local)
        self.rec_remote = self.server.ssh.normpath(self.rec_remote) 
        files.append([rec_local, self.rec_remote])
        if self.recinfo['is_flexible']:
            flex_local = self.recinfo['flex_res_file']
            self.flex_remote = self.vsdir + '/' + os.path.basename(flex_local)
            self.flex_remote = self.server.ssh.normpath(self.flex_remote)
            files.append([flex_local, self.flex_remote])
        problems = self.server.ssh.putfiles(files)
        if len(problems):
            self.dprint("errors copying receptor/flex files:", problems)
            return False
        return True


    def retrieveligands(self, normpath=True):
        """ get all ligands from ligand sources"""
        ligs = []
        for source in self.ligsource:
            ligs += [ self.server.ssh.normpath(x) for x in source.getItems() ]
        return ligs

    def makeliglist(self):
        """ generate the ligand index file parse by the job array"""
        source_ligs = self.retrieveligands()
        total = len(source_ligs)
        count = 0
        step = self.options['split']
        dest_ligs = []
        for lig in source_ligs:
            lig_name = hf.simplebasename(lig)
            actual_fname = '/'.join([ hf.splitdir(count, total, step), lig_name])
            #dest_ligs.append(vsdir + '/' + actual_fname) # LONGER SAFER
            dest_ligs.append(actual_fname) # SHORTER COMPACT
            count += 1
        seed_data = [ '%s:%s' % (s,d) for s,d in zip(source_ligs, dest_ligs) ]
        seed_data = '\n'.join(seed_data) + '\n'
        seed_file = self.vsdir + '/'+ self.SEEDFILE 
        if not self.server.writetext(seed_file, seed_data) == True:
            print "FATALERROR INDEX FILE"
            return False
        return True


    def makeconf(self):
        """ generate the config for docking including :
            - receptor name
            - flex_res name (if any)
            - cpu to use (from service settings)

        The engine.VinaGenConf is used in raw mode to modify files path to point
        to the fullpath, avoiding to copy rec+flex in each ligand directory
        """
        self.engine.vina_settings['cpu'] = self.service.config['multithread']
        config_text = self.engine.VinaGenConf(receptor = self.recname, raw=True)
        config_text['receptor'] = self.rec_remote
        if self.recinfo['is_flexible']:
            config_text['flex'] = self.flex_remote
        config_text = [ "%s = %s" % (k,v) for k,v in config_text.items() ]
        config_text = "\n".join(config_text) + "\n"
        self.config_file = self.vsdir + '/' + self.CONFIGFILE
        if not self.server.writetext(self.config_file, config_text):
            return False
        return True


    def setSchedulerParms(self, parm={}):
        # it must be a 
        pass

    def makejobtext(self, settings):
        """ create the text of job file with the specified settings"""
        sched = self.server.getScheduler()
        if sched == 'pbs':
            template = self._pbs_job_template
        elif sched == 'sge':
            template = self._sge_job_template
        text = "\n".join(template) + "\n"
        text = text % settings
        return text



    def generatejobs(self):
        """ generate the job script text file(s), depending on the max_job_arrays allowed"""
        self.jobs = []
        job_settings = { 'j_start'    : None, # |_ managed by the generation loop 
                         'j_end'      : None, # | 
                         'multithread' : self.service.config['multithread'],
                         'walltime'   : self.scheduler_parms['walltime'],
                         'queue'      : self.scheduler_parms['queue'],
                         'logfile'    : self.LOGFILE,
                         'vsdir'      : self.vsdir,
                         'seedfile'   : self.SEEDFILE,
                         'masterlog'  : self.MASTERLOG,
                         'status'     : self.STATUS,
                         'config'     : self.config_file,
                         'binary'     : self.server.ssh.normpath(self.service.config['command']),
                         'preprocess' : self.preprocess,
                         'postprocess': self.postprocess,
                       }
        # calculate how many separate jobs must be submitted (JOB ARRAY LIMIT)
        self.total_ligands = len(self.retrieveligands())
        #sched_info = self.server.properties['system']['scheduler_info'] 
        #max_jarrays = sched_info.get('max_job_array_size', total_ligands)
        max_jarrays = self.server.getMaxJobsPerArray()
        if max_jarrays == None:
            max_jarrays = self.total_ligands
        if max_jarrays >= self.total_ligands:
            jobs_count = 1
            max_jarrays = self.total_ligands
        else:
            jobs_count = self.total_ligands / max_jarrays
            if self.total_ligands % max_jarrays > 0:
                jobs_count += 1 
        for c in range(jobs_count): 
            j_start = 1 + max_jarrays * c
            j_end   = max_jarrays + max_jarrays * c
            if j_end > self.total_ligands:
                j_end = self.total_ligands
            job_settings['j_start'] = j_start
            job_settings['j_end'] = j_end
            if jobs_count > 1:
                job_settings['logfile'] = self.LOGFILE % c
            else:
                job_settings['logfile'] = self.LOGFILE % ""
            text = self.makejobtext(job_settings)
            self.jobs.append(text)
        self.status_file = self.vsdir + '/' + self.SEEDFILE
        return True

    def generateCollector(self):
        """ generate the script file for the collector job that runs tar on the list of dirs from self.STATUS
            to generate the final tar.gz file to be transferred

            possibly this job can be used to go on the nodes and recollect
            all the data to be moved back to homes if necessary
            (i.e. home not visible from nodes)
        """
        name = 'collector.j'
        self._collector_sge_template = [
            """#!/bin/bash""",
            """#$ -S /bin/bash""",
            """#$ -l h_rt=%(walltime)s""",
            """#$ -wd %(vsdir)s""",
            """#$ -o %(vsdir)s/%(logfile)s""",
            """#$ -j y""",
            """#$ -hold_jid %(dependency)s""",
            """# input variables""",
            #"""VSNAME='%(vsname)s'""",
            """STATUS='%(vsdir)s/%(status)s'""",
            """if [ ! -e $STATUS ]""",
            """  then""",
            """    echo "LOG FILE [$STATUS] DOES NOT EXIST!" 1>&2""",
            """    exit 1""",
            """fi""",
            """touch %(vsdir)s/.collecting""",
            #"""tar zcf %(vsname)s.tar.gz -T $STATUS %(receptor)s %(config)s""",
            """tar zcf %(vsname)s.tar.gz -T $STATUS %(receptor)s %(config)s %(seedfile)s""",
            """echo "tar file generated" >> %(vsdir)s/.collected""",
            """rm %(vsdir)s/.collecting""",
            ]

        self._collector_pbs_template = [
            """#!/bin/bash""",
            """#PBS -S /bin/bash """,                   # set the shell
            """#PBS -l nodes=1:ppn=1""",
            """#PBS -l walltime=%(walltime)s""",        # xx:xx:xx
            """#PBS -d %(vsdir)s""",                    # initial array directory
            """#PBS -o %(vsdir)s/%(logfile)s""",        # log file
            """#PBS -j oe """,                          # combine output and error logs in the same file
            """#PBS -W depend=%(dependency)s""",             # jobs to be waited for completion
            """VSNAME='%(vsname)s'""",
            """STATUS='%(vsdir)s/%(status)s'""",
            """if [ ! -e $STATUS ]""",
            """  then""",
            """    echo "LOG FILE [$STATUS] DOES NOT EXIST!" 1>&2""",
            """    exit 1""",
            """fi""",
            """touch %(vsdir)s/.collecting""",
            """tar zcf %(vsname)s.tar.gz -T $STATUS %(receptor)s %(config)s %(seedfile)s""",
            """echo "tar file generated" >> %(vsdir)s/.collected""",
            """rm %(vsdir)s/.collecting""",
            ]
        sched = self.server.getScheduler()
        if sched == 'pbs':
            template = self._collector_pbs_template
            cmd = """qsub %s"""
            #dependency = "afterany:"+ ",".join(self.job_id)
            dependency = "afteranyarray:"+ ",".join(self.job_id)
        elif sched == 'sge':
            template = self._collector_sge_template
            cmd = """qsub -terse %s"""
            #print "JOBID", self.job_id
            job_id = [ x.split(".", 1)[0] for x in self.job_id]
            #print "JOBID POST",job_id
            dependency = ",".join(job_id)

        data = { 'walltime'  : self.scheduler_parms['walltime'],
                 'vsdir'     : self.vsdir,
                 'logfile'   : 'collector_issues.log',
                 'vsname'    : self.name,
                 'dependency': dependency,
                 'status'    : self.STATUS,
                 'receptor'  : self.rec_remote.rsplit('/',1)[1],
                 'config'    : self.config_file.rsplit('/',1)[1],
                 'seedfile'  : self.SEEDFILE,
                }
        # generate text
        text = "\n".join(template) + "\n"
        text = text % data
        fname = self.vsdir + '/' + name
        cmd = cmd % fname
        if not self.server.writetext(fname, text):
            self.dprint("error writing text file")
            return False
        # submit with dependency from run jobs
        report = self.server.ssh.execmd(cmd)
        if report == False:
            self.dprint("error submitting the job script")
            return False
        stdout, stderr = report
        if len(stderr):
            print "SUBMISSION ERROR"
            print stderr
            return stderr
        self.job_id.append(stdout[0])
        return True


    def submit(self):
        """ write the job script files
            perform the submission on the server
            return the submission info (for the hosting app for a local copy)
        """
        submission_data = { 'error': None, 'info' : None }
        sched = self.server.getScheduler()
        for j in range(len(self.jobs)):
            text = self.jobs[j]
            if len(self.jobs) > 1:
                fname = self.JOBFILE % j
            else:
                fname = self.JOBFILE % ""
            # write job script
            fname = self.vsdir + '/' + fname
            if not self.server.writetext(fname, text):
                self.dprint("error writing text file")
                return False
            if sched == 'pbs':
                cmd = 'qsub %s' % fname
            elif sched == 'sge':
                cmd = 'qsub -terse %s' % fname
            report = self.server.ssh.execmd(cmd)
            if report == False:
                self.dprint("error executing the submission")
                return False
            stdout, stderr = report
            if len(stderr):
                self.dprint("submission returned errors", stderr)
                submission_data['error']  = stderr
                return submission_data
            job_id = stdout[0]
            if sched == 'sge':
                job_id = job_id.split(".", 1)[0]
            self.job_id.append(job_id)
        prj = self.jobinfo['prj']
        exp = self.jobinfo['exp']
        collector = self.generateCollector()
        if not collector == True:
            submission_data['error'] = 'Error in generating the collector [%s]' % collector

        info =  { prj : { exp : {  self.name : {  'type' : 'vs',
                                                  'date' : self.server.date(),
                                                  'hostname' : self.server.hostname(),
                                                  'servername': self.server.name(),
                                                  'status': 'submitted',
                                                  'statusfile' : self.STATUS,
                                                  'masterlog'  : self.MASTERLOG,
                                                  'jobid' : self.job_id,
                                                  'vsdir' : self.vsdir,
                                                  'total' : self.total_ligands,
                                                  'downloaded' : False,
                                                  'engine' : 'vina',
                                                  'resource' : 'ssh', } } } }  
        submission_data['info'] = info
        print "SUBMISSION INFO", info
        self.server.registerJob(info)
        return submission_data


class SGEgenerator:
    """provides the submit() method for the
        generation of SGE-compliant jobs
    """
    # http://wiki.ibest.uidaho.edu/index.php/Tutorials:_SGE_PBS_Converting
    # XXX CONVERSION PBS TO SGE
    def __init__(self):
        pass

#class VinaPBS(VinaPBSgenerator, VinaRunner):
#
#    pass

"""

class VinaSGE(SGEgenerator, VinaRunner):
    pass

class AutoDockPBS(PBSgenerator, AutoDockRunner):
    pass
    
class AutoDockSGE(SGEgenerator, AutoDockRunner):
    pass
""" 

    

class DockPostProcessing(RaccoonService):
    """
    DISABLED USELESS FOR NOW
    """
    
    def __init__(self, server, lig, rec, debug=False):
        RaccoonService.__init__(self, server, debug)
        

        self._srvtype = 'postprocess'
        self.lig = lig
        self.rec = rec

        self._requiredkw += ['command']
        #self._legal_values.update({'dockingengine': ['vina', 'autodock']})



        self._config_template = [
            '############# Raccoon service config file #################',
            '# NAME',
            '# The name of the service as it will be shown in the services list',
            'name = "%(name)s"',
            '\n# COMMENT',
            '# contains a one line description of the service',
            '# this is optional, so leave emtpy quotes if nothing is used',
            'comment = "%(comment)s"',
            '\n# SERVICE TYPE',
            '# define the calculation/data type provided',
            'service = postprocess',
            '\n# DATA SOURCE',
            '# specify which kind of data is required for the processing',
            '# allowed values are : ligand, receptor, both',
            'engine = %(engine)s',
            '\n# COMMAND NAME',
            '# command that is going to be used to run the docking',
            '# if full path is specified, it must be accessible from all nodes',
            'command = "%(command)s"',
            ]




##################################################################################
##################################################################################
####################   LIBRARY SECTION BELOW #####################################
##################################################################################
##################################################################################



class LibraryHostService(RaccoonService):
    """
        service that host libraries of structural data
        for docking outside the current server (i.e. providing remote
        access of libraries to other servers)
    """
    # XXX IT MUST PROVIDE A DOWNLOAD MECHANISM!
    # XXX remote_tar, transfer, local_untar?

    def __init__(self, server, config={}, debug=False):
        RaccoonService.__init__(self, server, debug)
        # default config file structure for the library server
        self._config_template = [
            '############# Raccoon Library Host config file #################',
            '# ',
            '# This file defines the configuration of a Raccoon Docking Library',
            '# hosting service.',
            '# ',
            '# A library is a homogeneus collection of structural files (ligands,',
            '# receptors, ...) that can be used for molecular mechanics',
            '# simulations.',
            '# ',
            '# NAME',
            '# The name of the library service (i.e. "commercially available compounds"',
            '# as it will be shown in the services list',
            'name = "%(name)s"',
            '\n# COMMENT',
            '# contains a one line description of the service',
            '# this is optional, so leave emtpy quotes if nothing is used',
            'comment = "%(comment)s"',
            '\n# SERVICE TYPE',
            '# define the calculation/data type provided'
            '# allowed values could: docking, md, filtering ...',
            'service = library_host',
            '\n# LIBRARIES',
            '# list of libraries that are served to clients',
            '# library names must be in the $RACDIR/library/index.lib',
            '# where the info on their files and properties is stored',
            '# for example:',
            '# lib = "nci_div2"',
            '#',
            '# to host automatically all libraries installed on the server',
            '# set "lib" to "_all_"',
            'lib = "_all_"',
            ]
        self._srvtype = 'library_host'
        self.config['service'] = self._srvtype
        self.config.update({ 'lib':[]} )
        self._requiredkw += ['service', 'lib']
        #self._legal_values.update( { 'engine': ['autodock', 'vina'] ,
        #             'multithread' : int } )
        if config:
            self.initService(config)


    def initService(self, config={}):
        """ initialize the service from the config"""
        self.dprint("attempt to initialize service [%s]" % config['name'], args=1,new=1)
        self.ready = False
        # config validation
        self.validateConfig(config)
        if not self.config['validated']:
            self.dprint("Service *FAILED* config validation")
            return False
        return self.validateLibNames()


    def validateLibNames(self, names=[]):
        """ check that names requested for the hosting
            are known to the server, i.e. listed in 
            the library index file
        """
        if names == []:
            names = self.config['lib']
        registered_libs = self.server.getLibraryNames()
        if '_all_' in names:
            self.dprint("all libraries requested ['_all_' kw]")
            if len(registered_libs) > 0:
                self.dprint("there is at least a library installed on the server")
                return True
            else:
                self.dprint("no libraries registered on the server... what to do? returning FALSE for now")
                return False
        else:
            for n in names:
                if not n in registered_libs:
                    return False
            return True


    def makeConfig(self):
        """generate a config file from the library service template
           populating the lib entries
        """
        template = self._config_template[:]
        # lib keyword in the config file can be specified one or more times
        if type(self.config['lib']) == type([]):
            lib_list = self.config['lib']
        else:
            lib_list = [ self.config['lib'] ]
        for lib in lib_list:
            l = 'lib = "%s"' % lib
            template.append(l)
        conf = self._makeConfigText(template)
        return conf

    def _checkLibraryFiles(self):
        """ maybe this should become a RacconService parent class
            method
        """
        pass


    def checkLibraryNodeAccessibility(self, libname):
        """
        depending on the scheduler/resmanager do this operation?

        NO! THIS IS RISKY! IF THE QUEUE IS FULL THE TIME IS INFINITE!
         - qsub -I
         - stat $theFirstFileOfTheLibrary
         - if it's ok, the file is accessible from the nodes

        """
        pass


class LibraryUploadService(RaccoonService):
    """
        service that allow to upload libraries of structural data
        for docking
    """
    def __init__(self, server, config={}, debug=False):
        RaccoonService.__init__(self, server, debug)
        # default config file structure for the library server
        self._config_template = [
            '############# Raccoon Library Upload config file #################',
            '# ',
            '# This file defines the configuration of a Raccoon Docking Library',
            '# upload service.',
            '# ',
            '# A library is a homogeneus collection of structural files (ligands,',
            '# receptors, ...) that can be used for molecular mechanics',
            '# simulations.',
            '# ',
            '# NAME',
            '# The name of the library as it will be shown in the services list',
            'name = "%(name)s"',
            '\n# COMMENT',
            '# contains a one line description of the service',
            '# this is optional, so leave emtpy quotes if nothing is used',
            'comment = "%(comment)s"',
            '\n# SERVICE TYPE',
            '# define the calculation/data type provided'
            '# allowed values could: docking, md, filtering ...',
            'service = library_upload',
            '\n# LIBRARY STORAGE PATH',
            '# define where the files are going to be updloaded on the server'
            '# allowed values could: docking, md, filtering ...',
            'location = library_upload',
            '\n# LIMITATIONS',
            '# limitations in library properties (i.e. number of files,',
            '# max_items = [ leave it empty if not enforced ]',
            'max_items =  ',
            ]
        self._srvtype = 'library_upload'
        self.config.update({'location':'', 'max_items':''})
        self._requiredkw += ['service', 'location' ]
        if config:
            self.initService(config)


    def uploadLibrary(self, info={}):
        """ upload the library in the path set in 'location'
            - read the self.config['location'] path
                - if the value is '' (empty) then use the $LIBRARY location
            - copy the files in that location
            - add the entry in the $LIBRARY/index.lib file

        info = { 'source' : '/local/path/to/lib_dir/',
                   'name'   : 'ThisIsTheLibraryNameOnTheServer',
                   'items'  : -1,
                   'destination': '/remote/path/on/the/server/where/to/write',
                   }
        """
        # path where to save the libraries
        if config.haskey('destination'):
            path = config['destination']
        else:
            path = self.server.ssh.getLibraryPath()

        if self._checkpathwritable( config['']):
            return
        # check the path is writable
        # check enough disk space?
        # check the name has not been used already
        # count that items match the maximum number of items allowed
        pass


class RemoteResources(RaccoonService):
    """ create a service that allow other services hosted 
        on this server to access remote files
        (i.e. if this service is active together with a docking
        service, it will allow it to get remote libraries or
        write the results into another server
    """
    def __init__(self, server, config = {}, debug=False):
        RaccoonService.__init__(self, server, debug)
        # USE WGET? SCP?
        # REQUIRE KEYS?
        from RaccoonSshTools import RaccoonServer, RaccoonSshClient
        #self.remoteserver = RaccoonServer()
        self.ssh = RaccoonSshClient()
        self.config.update( { 'server' : None, 
                            })
        # TODO INCOMPLETE

class Library:
    """ object to manage library properties

    """
    def __init__(self, server, libinfo = {}):
        self.server = server
        self.name = libinfo['name']
        self._type = libinfo['type']
        self.path = libinfo['path']
        self.date = libinfo['date']
        self.properties = {}

    def addLibToIndex(self, fname, force=False):
        """ add library info to an index file
            if a library with same name and type already exists,
            data will not be written, unless 'force=True'
        """
        libindex = self.server.ssh.readJson(fname)


class LigandLibrary(Library):
    """ object to manage library properties (LIGAND)
    """
    def __init__(self, server, libinfo={}):
        RemoteLibrary(self.server, libinfo)



    def updateProperties(self):
        """ populate and update properties of the
            ligand library
        """ 
        self.properties = {}
        for p in ['heavy', 'mw', 'tors', 'hbacc', 'hbdon']:
            try:
                self.properties[p] = libinfo[p]
            except:
                self.properties[p] = None



knownservices = { 'docking': DockingService, # AutoDock/Vina docking service
                  'library_host': LibraryHostService, # Library hosting service
                  'library_upload': LibraryUploadService, # Library hosting service
                  # XXX generic class to create?
                  'remoteresources' : RemoteResources, # Access to remote locations (read/write)
                }







about = "The MD Analysis workflows provide Molecular dynamics tools:\n"
about = about + "   AMDNBCR - AMD-based calculation \n"
about = about + "   GROMOS - trajectory clustering using GROMOS\n"
about = about + "   TrajQR - trajectory clustering using TrajQR\n"
about = about + "   NAMD_TG - NAMD-based calculation using TeraGRid resources\n"
about = about + "   NAMD_NBCR - NAMD-based calculation using NBCR resources\n"



#!/usr/bin/env python

from distutils.core import setup
from distutils.command.sdist import sdist
from distutils.command.install_data import install_data
from distutils.command.build_scripts import build_scripts
from glob import glob
import os

########################################################################
# Had to overwrite the prunefile_list method of sdist to not
# remove automatically the RCS/CVS directory from the distribution.
########################################################################

class modified_sdist(sdist):
    def prune_file_list(self):
        """
        Prune off branches that might slip into the file list as created
        by 'read_template()', but really don't belong there:
          * the build tree (typically 'build')
          * the release tree itself (only an issue if we ran 'sdist
            previously with --keep-temp, or it aborted)
        """
        build = self.get_finalized_command('build')
        base_dir = self.distribution.get_fullname()
        self.filelist.exclude_pattern(None, prefix=build.build_base)
        self.filelist.exclude_pattern(None, prefix=base_dir)

class modified_install_data(install_data):

    def run(self):
        install_cmd = self.get_finalized_command('install')
        self.install_dir = getattr(install_cmd, 'install_lib')
        return install_data.run(self)
    
class modified_build_scripts(build_scripts):
    
    def copy_scripts(self):
        #print "scripts:" , self.scripts

        for script in self.scripts:
            fnew = open(script, "w")
            forig = open(script+".py", "r")
            txt = forig.readlines()
            forig.close()
            fnew.write("#!/usr/bin/env python2.5\n")
            fnew.writelines(txt)
            fnew.close()
        build_scripts.copy_scripts(self)

        
########################################################################
# list of the python packages to be included in this distribution.
# sdist doesn't go recursively into subpackages so they need to be
# explicitaly listed.
# From these packages only the python modules will be taken
packages = ['CADD',
            'CADD/workflows/docking',
            'CADD/workflows/MDanalysis' , #scripts
            'CADD/workflows/user',
            'CADD/workflows/virtualScreening',
            'CADD/workflows/visualization',
            'CADD/Raccoon',
            'CADD/Raccoon/icons',
            'CADD/Raccoon2',
            'CADD/Raccoon2/gui',
            ]

# list of the python modules which are not part of a package
py_modules = ['CADD/bin/runCADD',
              ]

# list of the files that are not python packages but are included in the
# distribution and need to be installed at the proper place  by distutils.
# The list in MANIFEST.in lists is needed for including those files in
# the distribution, data_files in setup.py is needed to install them
# at the right place.
data_files = []

def getDataFiles(file_list, directory, names):
    fs = []
    for name in names:
        ext = os.path.splitext(name)[1]
        #print directory, name, ext, len(ext)
        if ext !=".py" and ext !=".pyc":
            fullname = os.path.join(directory,name)
            if not os.path.isdir(fullname):
                fs.append(fullname)
    if len(fs):
        file_list.append((directory, fs))

os.path.walk("CADD", getDataFiles, data_files)

scripts = ['CADD/bin/runCADD']
# description of what is going to be included in the distribution and
# installed.
from version import VERSION
setup(name = 'CADD',
      version = VERSION,
      description = "CADD - A computer aided drug design using Vision",
      author = 'UCSD',
      author_email = 'wili@ucsd.edu',
      download_url = 'http://nbcr.sdsc.edu/pub/wiki/index.php?title=CADD_Pipeline',
      url = 'http://nbcr.sdsc.edu/pub/wiki/index.php?title=CADD_Pipeline',
      packages = packages,
      py_modules = py_modules,
      data_files = data_files,
      scripts = scripts,
      cmdclass = {'sdist': modified_sdist,
                  'install_data': modified_install_data,
                  'build_scripts': modified_build_scripts
                  },
      )



#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


import paramiko as pmk
import socket
#import os
import sys
import time
from datetime import datetime
import errno
import Queue
import threading
import tarfile
import os
from DebugTools import DebugObj
import sys
from CADD.Raccoon2 import HelperFunctionsN3P as hf

# NOTES
# http://jessenoller.com/2009/02/05/ssh-programming-with-paramiko-completely-different/


# when transferring files, spawn a thread by opening a connection
# to the server and letting it run on the background...



class RaccoonSshClient(DebugObj):
    """provide the basic features for using a remote SSH server
       (i.e. linux cluster, remote machines, etc...)

        openconnection()    initialize the connection (self.connection)
        closenConnection()  close the connection (self.connection=None)
        execmd()           execute remote commands
        getfiles()          copy remote files locally
        putfiles()          copy local files remotely
        delfiles()          delete remote files/dirs
        openfile()          open a remote file (return pointer)
        listdir()           return the files present in the path
        

        NOTES: basic operation functions will normalize the path (self.normpath):
            - self.openfile
            - self.exists

    """
    def __init__(self, address=None, username=None, password=None, name='RaccoonSshClient_server', 
                debug=False, usesyshostkeys=True, autoaddmissingkeys=True, 
                pkey=None, autoconnect=False):
        
        DebugObj.__init__(self, debug)

        self.autoaddmissingkeys = autoaddmissingkeys
        self.usesyshostkeys = usesyshostkeys
        self.address = address
        self.username = username
        self.password = password
        self.name = name
        self.pkey = pkey
        self.connection = None
        self.connectiontime = None
        if autoconnect:
            self.openconnection()

        #pmk.util.log_to_file("paramiko_raccoon_connection.log")


    def setserver(server={}, autoconnect=False):
        # reset the server parameters...
        # FIXME complete or remove...
        print "\n\n\n************** WARNING! INCOMPLETE FUNCTION!!!! CALL FOR HELP!!! ****"
        print "THIS FUNCITION SHOULD BE REMOVED IN THE FINAL VERSION SINCE IT IS NOT USED"
        if not self.connection == None:
            self.closeconnection()
        keys = server.keys()
        if 'address' in keys:
            self.address = server['address']
        if 'username' in keys:
            self.username = server['username']
        if 'password' in keys:
            self.password = server['password']
        if 'pkey' in keys:
            self.pkey = server['pkey']
        if autoconnect:
            self.openconnection()

    def openconnection(self):
        """
        set the current server to be 'servername' and 
        initialize the connection (username, passwd)

        the connection start time is stored in self.connectiontime

        """
        if not self.connection == None:
            self.dprint('connection already open')
            return 1
        self.dprint("opening a new connection", new=1)
        conn = pmk.SSHClient()
        if self.autoaddmissingkeys: # 
            conn.set_missing_host_key_policy(pmk.AutoAddPolicy())

        # FIXME this should be configurable
        options = {'port': 22} 
        if self.usesyshostkeys:     # .ssh/authorized_keys
            try:
                self.dprint("using SYSTEM HOST KEYS")
                conn.load_system_host_keys()
            except IOError:
                e = sys.exc_info()[1]
                self.dprint("No keys found [ %s ]" % e)
                err = 'keys not available in [ %s ]' % e
                err_o = IOError
        elif not self.pkey == None:
            options['key_filename'] = self.pkey
            self.dprint("using user defined HOST KEY [%s]" % self.pkey)
        else:
            self.dprint("trying password [%s]" % self.password)
            options['password'] = self.password
        err = 'Connection not open (UNKNOWN ERROR)'
        shell = 'unk'
        shell_errors = ['none']
        try:
            conn.connect(hostname=self.address, username=self.username, **options) # password=self.password)
            # NOTE FRAGILE SHELL TEST 
            # this test is used for fragile shells (i.e. csh) to trigger 
            # SSHException(EOF) if there are errors in the rc files (i.e.
            # .cshrc)
            cmd = "echo $SHELL"
            stdin_raw, stdout_raw, stderr_raw = conn.exec_command(cmd)

            shell_errors = stderr_raw.readlines()
            shell = stdout_raw.readlines()[0].strip()
            stdin_raw.close()
            stdout_raw.close()
            stderr_raw.close()
            sftp = conn.open_sftp()
            sftp.close()
            # FRAGILE SHELL TEST /END
            self.connection = conn
            self.connectiontime = datetime.now()
            err = None
        except pmk.BadHostKeyException:
            """if the server's host key could not be verified"""
            err = "host key could not be verified (%s)" % sys.exc_info()[1]
            err_o = pmk.BadHostKeyException
        except pmk.AuthenticationException:
            """ if authentication failed"""
            err = "error in the authentication (%s)" % sys.exc_info()[1]
            err_o = pmk.AuthenticationException
        except pmk.SSHException:
            """ if there was any other error connecting or establishing an SSH session"""
            err = "error in establishing SSH session (%s)" % sys.exc_info()[1]
            if "EOF" in err:
                if len(shell_errors):
                    err += ('\n\nThe following errors have been reported '
                            'by the remote shell:\n\n'
                            '====[ %s ]====\n'
                            '\n%s\n================\n'
                            '\nConnection failure could be due to errors in setting '
                            'files (i.e. ".*rc", ".login" files) of the remote shell.'
                            ) % (shell, '\n'.join(shell_errors))
            err_o = pmk.SSHException
            
        except socket.error:
            """ a socket error occurred"""
            err = "socket error (%s)" % sys.exc_info()[1]
            err_o = socket.error
        except:
            """ undefined error"""
            err = "unknown error (%s)" % sys.exc_info()[1]
            err_o = None

        finally:
            self.dprint("SHELL [%s]" % shell)
            if len(shell_errors):
                for i in shell_errors:
                    self.dprint("SHELL ERROR> %s" %i)
            if err == None:
                self.dprint("[ connection open ]")
                self.expanduservar()
                return 1
            else:
                print "openconnection> an error occurred : ", err
                self.dprint("connection not opened, error[%s]" % err)
                return err, err_o

    def expanduservar(self):
        """ resolve the "~" expanding it to the user home fullpath"""
        self.dprint("set the uservar '~'...")
        cmd = 'echo ~'
        out, err = self.execmd(cmd)
        uservar = out[0]
        self.dprint("found [%s]")
        self.uservar = uservar


    def closeconnection(self):
        """ close connection and reset the connection time to None """
        self.dprint("closing open connection...", new=1)
        if not self.connection == None:
            self.connection.close()
            self.connection = None
            self.connectiontime = None
            self.dprint("closed.")
            self.uservar = None
            return
        self.dprint("no open connection to close...")


    def execmd(self, command, auto=False, getlines=True, dostrip=True, full = False, forceclose=False):
        """ execute command on remote server and return stdout and stderr
            as lists
        
            getlines: if true, lists instead of pointers are returned
            dostrip : (if getlines) lines are stripped
            full    : return stdin, stdout, stderr [default: stdout, stderr]
        """
        self.dprint('COMMAND:[%s]' % command, new=1)
        if self.connection == None:
            print "No connection"
            return False
        try:
            self.dprint("executing [%s]..." % command)
            stdin_raw, stdout_raw, stderr_raw = self.connection.exec_command(command)
            if forceclose:
                try:
                    print "Closing..."
                    stdin_raw.write('exit')
                    stdin_raw.close()
                except: 
                    print "CATCHED SOMETHING", sys.exc_info()[1]
                    pass
                self.dprint("DONE (force closing)")
                return  stdout_raw, stderr_raw 
            self.dprint("DONE")
        except:
            print "excmd> ERROR! [%s]" % (sys.exc_info()[1])
            return False
        if getlines and not forceclose:
            #stdin = stdin.readlines()
            stdout = stdout_raw.readlines()
            stderr = stderr_raw.readlines()
            stdout_raw.close()
            stderr_raw.close()
            if dostrip:
                #from string import strip
                stdout = [x.strip() for x in stdout]
                stderr = [x.strip() for x in stderr]
        if full:
            return (stdin_raw, stdout_raw, stderr_raw)
        return (stdout, stderr)

    def normpath(self, dirname):
        """ normalize the path by removing the user tildes if present 
           
            TODO: check if it is necessary to expand it? ("~" -> "/home/userx") ?

            fragile!?

            this should be safer:

            dirname = self.execmd("echo %s" % dirname, getlines=1)[0][0]
        """
        #return self.execmd("echo %s" % dirname, getlines=1)[0][0] # XXX VERY SLOW!
        if dirname.startswith("~/"):
            dirname = self.uservar+'/'+dirname[2:]
        return dirname


    def getfiles(self, flist=[], auto=True):
        """
        get fname from server 

        flist = [  ['source1', 'dest1'], 
                   ['source2', 'dest2'], 
                   ...
                ]
        """
        if self.connection == None and not auto:
            self.dprint(" no connection open on the server [%s]" % self.server_info['address'])
            return
        self.openconnection()
        sftp = self.connection.open_sftp()
        problems = []
        for f in flist:
            try:
                sftp.get(self.normpath(f[0]), self.normpath(f[1]))
            except:
                problems.append((f, sys.exc_info()[1]))
        sftp.close()
        return problems

        
    def putfiles(self, flist=[], mode=None, auto=True):
        """
        copy file on the server
        flist = [ [ source1, dest1], [source2, dest2], ... ]

        if optional 'mode' is provided, the file attributes will be
        changed as 'chmod mode filename'

        NOTE: if a remote directory is specified, it must exist!

        """

        if self.connection == None and not auto:
            print "putfiles> no connection open on the server [%s]" % self.server_info['address']
            return
        self.openconnection()
    
        sftp = self.connection.open_sftp()
        problems = []
        for f in flist:
            try:
                sftp.put(self.normpath(f[0]), self.normpath(f[1]))
            except:
                problems.append([f, sys.exc_info()[1]])
        sftp.close()

        if not mode == None:
            files = [ self.normpath(x[1]) for x in flist]
            files = " ".join(files)
            cmd = 'chmod %s %s' % (mode, files)
            self.execmd(cmd)
        return problems

    def makedir(self, dirname, mode=711, auto=True):
        """ probably not very efficient recursive hack to 
            allow parent creation when making a dir
        """
        # XXX
        # XXX what about executing the mkdir -p command and thazzit?
        # XXX

        dirname = self.normpath(dirname)
        cmd = """mkdir -m %s -p %s""" % (mode, dirname)
        cmd = """mkdir -p %s""" % ( dirname)
        report = self.execmd(cmd, getlines=True)
        if report == False:
            print "makedir> error in creating directory"
            return
        stdout, stderr = report
        self.dprint("STDOUT> %s" % stdout)
        self.dprint("STDERR> %s" % stderr)
        if len(stderr) == 0:
            self.dprint("no errors, success")
            return True
        else:
            self.dprint("ERRORS!")
            return stderr
        
        """
        forbidden = ['~', '/', None]
        dirname = self.normpath(dirname)
        if self.connection == None and not auto:
            print "makeDir> no connection open on the server [%s]" % self.server_info['address']
            return
        self.openconnection()

        sftp = self.connection.open_sftp()
        if '/' in dirname:
            parent = dirname.rsplit('/', 1)[0]
        else:
            parent = None
        if (not parent in forbidden) and (not self.exists(parent)):
            # XXX WARNING! FRAGILE HERE XXX #
            print "parent doesn't exists [%s]" % parent
            self.makedir(parent)
        sftp.mkdir(dirname, mode=mode)
        sftp.close()
        """

    def delfiles(self, flist=[], auto=True):
        """ delete files in the remote location 
            flist can contain both files and directories

            NOTE: better user self.execmd( 'rm -fr ...') ?
        
        """

        if self.connection == None and not auto:
            print "delfiles> no connection open on the server [%s]" % self.server_info['address']
            return
        self.openconnection()

        sftp = self.connection.open_sftp()
        problem = None
        flist = ' '.join(["%s" % x for x in flist ])
        try:
            cmd = 'rm -fr %s' % flist
            out, err = self.execmd(cmd)
        except:
            problem = "Error executing command. Cmd[%s], Error[%s]" % (cmd, sys.exc_info()[1])
        if len(err):
            problem = "Error executing command. Cmd[%s], Error[%s]" % (cmd, sys.exc_info()[1])
        sftp.close()
        return problem


    def openfile(self, fname, mode='r', bufsize=-1, auto=True):
        """
        open a remote file and 
        return the filepointer
        """
        fname = self.normpath(fname)
        self.dprint("Opening (normalized) filename [%s]" % fname)
        if self.connection == None:
            if not auto:
                print "openfile> no connection open on the server [%s]" % self.server_info['address']
                return
        else:
            self.openconnection()

        sftp = self.connection.open_sftp()
        try:
            return (sftp.open(fname, mode=mode, bufsize=bufsize), sftp)
        except:
            return sys.exc_info()[1]


    def touchfile(self, fname, auto=True):
        """
        touch a file on the remote server
        """
        fname = self.normpath(fname)
        out = self.openfile(fname, 'a')
        try:
            out[0].close()
            out[1].close()
            return True
        except:
            return out
            
    def remotereadl(self, fname, dostrip=1, noempty=1, removecomments=1):
        """ read a remote file and return lines
         
            by default, lines are stripped
        """
        from string import strip
        out = self.openfile(fname)
        if not type(out) == type(()):
            print "remotereadl> Error reading file [%s] : %s" % (fname, out)
            return False
        fp, sftp = out
        lines = fp.readlines()
        fp.close()
        sftp.close()
        if removecomments:
            #lines = [ l for l in lines if not l.startswith('#') ]
            # this removes comments more efficiently
            lines = [ l.split("#", 1)[0] for l in lines]
        if noempty:
            lines = [ l for l in lines if l.strip() ]
        if dostrip:
            lines = map(strip, lines)
        return lines

    def remotewritel(self, fname, _list, mode='a', addnl=1):
        """ write a list on a remote file
            by default new lines are added unless 'addnl=0' is specified
            also, data is appended to existing files unless
            mode='w' is specified
        """
        nl=''
        if addnl: nl='\n'
        out = self.openfile(fname)
        if not type(out) == type(()):
            print "remotewritel> Error opening file [%s, mode:%s] %s" % (fname, mode, out)
            return False
        fp, sftp = out
        for l in _list:
            fp.write( l+nl )
        fp.close()
        return True


    def exists(self, path, auto=True, info=False):
        """Return True if the remote path exists
        """
        path = self.normpath(path)
        if self.connection == None and not auto:
            print "openfile> no connection open on the server [%s]" % self.server_info['address']
            return
        self.openconnection()
        try:
            sftp = self.connection.open_sftp()
            x = sftp.stat(path)
            sftp.close()
            self.dprint("File [%s] exists: [%s]" % (path, x))
        except IOError, e:
            if e.errno == errno.ENOENT:
                self.dprint("File [%s] not found" % path)
                return False
            raise
        else:
            if info:
                prop = { 'mode' : x.st_mode,
                         'size' : x.st_size,
                         'uid'  : x.st_uid,
                         'gid'  : x.st_gid,
                         'atime': x.st_atime,
                         'mtime': x.st_mtime,
                        }
                return prop
            return True

    def findcommand(self, command, which=True):
        """ test if an executable command can be found with is 
            full-path or can be found by the UNIX command 'which'. 
        """
        found = False
        if self.exists(command):
            found = command
        if not found and which:
            self.dprint("which requested")
            cmd = '''which "%s"''' % command
            report = self.execmd(cmd)
            if report == False:
                return False
            stdout, stderr = report
            self.dprint("STDOUT[%s]" % stdout)
            self.dprint("STDERR[%s]" % stderr)
            if len(stdout):
                if not stdout[0] == '':
                    found = stdout[0]
                    if not ("not found" in found): 
                        # Rats! Csh, the shell from Hell returns
                        # the "command not found" as STDOUT...
                        self.dprint("Found binary => [%s]" % (stdout[0]))
                    else:
                        found = False
                else:
                    self.dprint("Impossible to find %s binary => [??]" % (command))
        #print "FINDCOMMAND>", found
        return found        


    def listdir(self, path):
        """ return the list of directories on the requested path """
        path = self.normpath(path)
        try:
            sftp = self.connection.open_sftp()
            ls = sftp.listdir(path)
            sftp.close()
            return ls
        except:
            return False

    def diskspace(self, path, human=True):
        """ check available disk space in the requested
            location

            by default the space is reported in
            human-readable format
        """
        suffixd = { 'P' : ' petabytes (a lot!)',
                    'T' : ' terabytes',
                    'M' : ' megabytes',
                    'K' : ' kilobytes',
                  }
        unk = ' unknown [furlong per forthnight?]'
        if human: hopt = '-h'
        else: hopt = ''
        cmd = """df %s %s | tail -1 | awk '{print $2}'""" % (hopt,path)
        report = self.execmd(cmd)
        if report == False:
            print "diskavail> error in the connection"
            return False
        stdout, stderr = report
        if len(stderr):
            print "diskavail> Errors encountered:\n%s\n" % "\n".join(stderr)
        if not len(stdout):
            print "diskavail> Empty stdout returned."
            return False
        space = stdout[0]
        if human:
            suffix = suffixd.get(space[-1], unk)
            space = space[:-1] + suffix
        return space


class ThreadedOperation: #(threading.Thread):
    """Basic class defining threaded operations

    it shuold be used to submit threads of given functions
    providing a queue mechanism for each one of them

    """
    def __init__(self, queue_mode = 'lifo'):
        #threading.Thread.__init__(self)
        if queue_mode == 'lifo':
            self.queue = Queue.LifoQueue
        elif queue_mode == 'fifo':
            self.queue = Queue.Queue

        self.jobs = []

    def run(self, func, args={}):
        self.pax
        pass


    def stop(self):
        self._STOP = False
        self._completed = False
        pass




class FileMule(DebugObj):
    """ Class to handle non-blocking transfers of large number of files/dirs
        such as ligand libraries with nested sub-dirs. It supports GZ and BZ2
        compressions.

        The transfer opens remote tar writing [or reading] pipes and transfer 
        the files through it, letting the remote tar process to create [or read]
        all files and dirs.
        The transfer mimics the following shell commands:

        WRITING:
            local> tar cf - $LOCAL_DIR | ssh "cd $REMOTE_DIR; tar xf -"

        READING:
            local> ssh "tar cf - $REMOTE_DIR" | tar xf -

        It can be spawned from a connection that is already open
        by passing the ssh client to it.
        
        It has an internal counter to be used to check status when it is 
        executed within its own thread (i.e. in a GUI):

        
            import threading
            ssh = RaccoonSshClient( address, username, password)
            transfer = FileUploader( parent = ssh, files_list, path='remote_path',
                tarname = 'temptar', compression = 'gz')

            thread = threading.Thread(target = transfer.upload)
            thread.start()

            status = transfer.progress(percent=True)


        
        NOTE: autoconnect is True by default (different than RaccoonSshclient
              default).
              
    """
    def __init__(self, ssh=None, address=None, username=None, password=None, name='RaccoonSshClient_server', 
            debug=False, usesyshostkeys=True, autoaddmissingkeys=True, pkey=None, autoconnect=True):
            ## ,  # XXX
            ## files=None, path = None, targetname = 'tmpfile', compression='gz'):

        # files, path, targetname and compression shuold be 
        # moved to the respective functions upload() and download()

        DebugObj.__init__(self, debug)

        if ssh:
            self.dprint("parent SSH connection provided: [%s]" % ssh)
            self.ssh = ssh
        else:
            self.dprint("open new SSH connection...")
            self.ssh = RaccoonSshClient(address, username, password, name, debug, 
                usesyshostkeys, autoaddmissingkeys, pkey, autoconnect)
            self.dprint("SSH connection open [%s]" % self.ssh)

        self._ready = False
        self._counter = 0.
        self._total = 0.
        self._STOP = False
        self._running = False
        self.pending = None
        self._status = { 'completed' : False, 'error': None }

        self.test()

    def test(self):
        """ check that requirements to transfer the files are satisfied"""
        self._ready=False
        if self._checkTar():
            self.dprint("tar test passed")
            self._ready = True
        else:
            self.dprint("test failed at tar.")

    def _checkTar(self):
        """check if the server has tar installed"""
        return self.ssh.findcommand("tar", which=True) 

    def stop(self):
        """ set the STOP variable to true, so running operations
            that check it can stop
        """
        if not self.checkPending():
            return
        msg = "transfer stop requested"
        self.dprint(msg)
        self._status['error'] = msg
        self._STOP = True
        self._completed = False

    def close(self):
        self._running = False
        self._STOP = False
        self.tarfile.close()
        self._status = {'completed' : False, 'error' : 'stopped'} 


    def upload(self, files=[], remotedestpath='.', compression='gz', bg=False): #, date=True):
        """ transfer files to a remote location

            this operation is designed to be threaded

            files = [ [ source_full_path_and_name, dest_relative_path_and_name ],
                   ]

            TODO investigate if files must be provided as pair (to avoid long paths
             and so on...)
        """
        if bg == False:
            self._upload_tar(files, remotedestpath, compression)
        else:
            self.pending = threading.Thread(target = self._upload_tar, args = ( files, remotedestpath, compression))
            self.pending.start()           

    def _upload_tar(self, files=[], remotedestpath='.', compression='gz'):
        # check if the destination path exists, otherwise exit..
        if remotedestpath == None:
            self.dprint("ERROR: Remote path not specified, returning")
            self._status = {'completed' : False, 'error' : 'remote path not specified'}
            return False
        if not self.ssh.exists(remotedestpath):
            self.dprint("Remote path [%s] do not exist, returning" % remotedestpath)
            self._status = {'completed' : False, 'error' : 'remote path does not exist'}
            return False
        self._status['completed'] = False
        self._counter = 0
        self._total = len(files)
        if compression == None :
            opt = ''
            mode = 'w'
        elif compression == 'gz' :
            opt = 'z'
            mode = 'w:gz'
        elif compression == 'bz2': 
            opt = 'j'
            mode = 'w:bz2'
        cmd = 'cd %s; tar %sxf -' % (remotedestpath, opt)
        #cmd = 'cd %s; tar %scf trap.tar' % (remotedestpath, opt)
        self._running = True
        self.dprint("command [%s]" % cmd)
        stdin, stdout, stderr = self.ssh.execmd(cmd, getlines=False, full=True)
        self.tarfile = tarfile.open(fileobj = stdin, mode=mode)
        for f in files:
            if self._STOP:
                self.dprint("Stop requested, halting")
                self.close()
                return
            infile, destfile = f
            destfile = self.ssh.normpath(destfile)
            self.dprint("Remote TAR writing src[%s] => dest[%s] %d" % (infile, destfile, self.progress(percent=1)) ),
            self.tarfile.add(infile, arcname=destfile)
            #self.tarfile.fileobj.flush()
            self._counter += 1
        self.tarfile.close()
        stdin.flush()
        # http://stackoverflow.com/questions/8052840/paramiko-piping-blocks-forever-on-read
        stdin.channel.shutdown_write()
        stdin.close()
        stdout.close()
        stderr.close()
        self._status = {'completed' : True, 'error' : None}
        self._running = False
        ### if date:
        ###     print "WRITE A DATE COMMAND LOG IN A FILE !!!"
        return True

    def checkPending(self):
        """check if a pending operation is pending"""
        if self.pending == None:
            return False
        pending = self.pending.isAlive()
        if not pending:
            self.pending = None
        return pending


    def downloadBigFile(self, source, destpath='.', bg=True, cb=None):
        """ manage the download of one single big result (tar) file """
        info = self.ssh.exists(source, info=True)
        if info == False:
            self.dprint("non-existent file [%s]", source)
            return False
        self._total = float(info['size']) # the size of the final file
        localname = source.rsplit("/", 1)[1]
        self._localfile = "%s%s%s" % ( destpath, os.sep, localname)
        self.download([source], destpath, compression=None, bg=bg, cb=cb)
        

    def updateDownloadBigFile(self):
        """ update download status of big files transfers"""
        if self.checkPending() == False:
            return 100.00
        #self._counter
        try:
            self.dprint("checking local size of [%s]" % self._localfile)
            curr_size = os.stat(self._localfile).st_size
            self.dprint(' found %s' % curr_size)
        except:
            self.dprint('file checking error [%s]' % sys.exc_info()[1] )
            curr_size = 0
        pc = hf.percent(curr_size, self._total)
        self.dprint("current size [%d/%d] : %2.3f%%" % (curr_size, self._total, pc) )
        return pc

    def download(self, files=[], localdestpath='.', compression='gz', bg=False, cb=None):
        """ transfer files from a remote location

            this operation is designed to be threaded

            files = [ [ source_full_path_and_name, dest_relative_path_and_name ],
                   ]

            TODO investigate if files must be provided as pair (to avoid long paths
             and so on...)
        """
        args = (files, localdestpath, compression, cb)
        if bg == False:
            #self._download_tar(files, localdestpath, compression, cb)
            self._download_tar(*args)
        else:
            #self.pending = threading.Thread(target = self._download_tar, args = ( files, localdestpath, compression, cb))
            self.pending = threading.Thread(target = self._download_tar, args = args)
            self.pending.start() 



    def _download_tar(self, source, localdestpath='.', compression='gz', bg=False, cb=None ):
        """ download files using the tar file
            no progress will be available when downloading
            how to handle dirs versus files?

            /garihome/forli/libs/catrame/
        """
        self._status = {'completed' : False, 'error' : None} 
        # XXX TO DO ADD HELPER PRE_FUNCTION TO CREATE BG THREAD...
        if not os.path.isdir(localdestpath):
            self.dprint("ERROR: local destination path is not a dir [%s], returning" % localdestpath)
            self._status['error'] = 'localdestpath is not a dir'
            return
        os.chdir(localdestpath)

        self._problematic = []
        self._status['completed'] = False
        self._counter = 0 # No progress is available
        # test if the remote path exists
        if not self._ready:
            self.dprint("Transfer not ready. Returning...")
            if not cb == None:
                cb("transfer not ready")
            return -1

        # find which compression option is requred
        #if compression == None :
        #    opt = ''
        #    mode = 'r'
        #elif compression == 'gz' :
        #    opt = 'z'
        #    mode = 'r:gz'
        #elif compression == 'bz2': 
        #    opt = 'j'
        #    mode = 'r:bz2'
        #
        #self.dprint('\n\n\n\nTar mode is [%s], opt[%s]' % (mode, opt))
        self._running = True
        #print "RaccoonSshTools.py > _download_tar"
        for fname in source:
            if not self.ssh.exists(fname):
                self._problematic.append([fname, 'not existent file'])
                self.dprint("the file [%s] does not exist, skipping..." % f)
                if not cb == None: cb("non-existent file [%s]" % fname)
                continue            # files must be stored as fullpath?
            #if f[-1] == '/': 
            #    self.dprint("the request was about a dir, probably [%s]" % f)
            #    f = f[:-1] # it's a dir
            #f = f.rsplit('/', 1)
            #if len(f) == 2:
            #    dname = f[0]
            #    fname = f[1]
            #else:
            #    cd = "."
            #    fname = f[0]
            #
            try:
            #if 1:
                if self._STOP:
                    self.dprint("Stop requested, halting")
                    self.close()
                    return
                #cmd = 'cd %s; tar %scf - "%s"' % (dname, opt, fname)
                #self.dprint("executing remote command [%s]" % cmd)
                #self.stdin, self.stdout, self.stderr = self.ssh.execmd(cmd, getlines=False, full=True)
                #print "STDIN ", self.stdin
                #print "STDOUT", self.stdout
                #print "STDERR", self.stderr
                #print "HERE THE ERROR SHOULD SHOW UP"
                #print "OPENSSHFILE", f
                fp, sftp = self.ssh.openfile(fname)
                #print "TARFILE", 
                self.tarfile = tarfile.open(fileobj = fp)
                #print "THERE WERE NO ERRORS?"
                self.tarfile.extractall()
                self._counter += 1
                self.tarfile.close()
                #self.stdin.close()
                #self.stdout.close()
                #self.stderr.close()
                fp.close()
                sftp.close()
            #else:
            except:
                #print "ERROR SOMEWHERE HERE", sys.exc_info()[1]
                if not cb == None: cb("eror: %s" % sys.exc_info()[1])
                self._problematic.append([fname, 'not existent file'])

        self._running = False
        self._status = {'completed' : True, 'error' : None}        
        if not cb == None: cb('completed')





    def progress(self, percent=False):
        """ available only when uploading by counting the
            input files processed
        """
        if percent:
            try:
                return (float(self._counter)/self._total)*100.
            except ZeroDivisionError:
                return 0.0
                
        return self._counter




        



